"""Built-in 4-vector classes."""

__all__: tuple[str, ...] = ()

from typing import TYPE_CHECKING, Any, cast

import jax
from jaxtyping import Array, Bool
from quax import register

import quaxed.numpy as jnp
from dataclassish import replace

from .spacetime import FourVector

if TYPE_CHECKING:
    import coordinax  # noqa: ICN001


@register(jax.lax.add_p)
def add_p_4vs(self: FourVector, other: FourVector, /) -> FourVector:
    """Add two 4-vectors.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> w1 = cx.FourVector (t=u.Quantity(1, "s"), q=u.Quantity([1, 2, 3], "m"))
    >>> w2 = cx.FourVector (t=u.Quantity(2, "s"), q=u.Quantity([4, 5, 6], "m"))
    >>> w3 = w1 + w2
    >>> print(w3)
    <FourVector: (t[s], q=(x, y, z) [m])
        [3 5 7 9]>

    """
    return replace(
        self,
        t=self.t + other.t,
        q=cast("coordinax.vecs.AbstractPos3D", self.q + other.q),
    )


@register(jax.lax.broadcast_in_dim_p)
def broadcast_in_dim_p_4v(
    obj: FourVector, /, *, shape: tuple[int, ...], **kwargs: Any
) -> FourVector:
    """Broadcast in a dimension."""
    batch = shape[:-1]
    return replace(
        obj,
        t=jnp.broadcast_to(obj.t, batch),
        q=cast("coordinax.vecs.AbstractPos3D", jnp.broadcast_to(obj.q, (*batch, 3))),
    )


@register(jax.lax.eq_p)
def _eq_4v_4v(lhs: FourVector, rhs: FourVector, /) -> Bool[Array, "..."]:
    """Element-wise equality of two 4-vectors.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> w1 = cx.FourVector (t=u.Quantity([1, 2], "s"),
    ...                    q=u.Quantity([[1, 2, 3], [4, 5, 6]], "m"))
    >>> w2 = cx.FourVector (t=u.Quantity(1, "s"), q=u.Quantity([1, 3, 3], "m"))

    >>> w1 == w2
    Array([False, False], dtype=bool)

    """
    return jnp.logical_and(jnp.equal(lhs.q, rhs.q), jnp.equal(lhs.t, rhs.t))


@register(jax.lax.neg_p)
def neg_p_4v(self: FourVector, /) -> FourVector:
    """Negate the 4-vector.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> w = cx.FourVector (t=u.Quantity(1, "s"), q=u.Quantity([1, 2, 3], "m"))
    >>> print(-w)
    <FourVector: (t[s], q=(x, y, z) [m])
        [-1 -1 -2 -3]>

    """
    return replace(self, t=-self.t, q=cast("coordinax.vecs.AbstractPos3D", -self.q))


@register(jax.lax.sub_p)
def sub_p_4v_4v(lhs: FourVector, rhs: FourVector, /) -> FourVector:
    """Add two 4-vectors.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> w1 = cx.FourVector (t=u.Quantity(1, "s"), q=u.Quantity([1, 2, 3], "m"))
    >>> w2 = cx.FourVector (t=u.Quantity(2, "s"), q=u.Quantity([4, 5, 6], "m"))
    >>> w3 = w1 - w2
    >>> print(w3)
    <FourVector: (t[s], q=(x, y, z) [m])
        [-1 -3 -3 -3]>

    """
    return replace(
        lhs, t=lhs.t - rhs.t, q=cast("coordinax.vecs.AbstractPos3D", lhs.q - rhs.q)
    )



"""Register primitives."""

__all__: tuple[str, ...] = ()


import jax
from quax import register

from dataclassish import replace

from .coordinate import Coordinate


@register(jax.lax.neg_p)
def neg_p_coord(x: Coordinate, /) -> Coordinate:
    """Negate a coordinate.

    Examples
    --------
    >>> import coordinax as cx

    >>> data = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
    >>> coord = cx.Coordinate(data, cx.frames.ICRS())

    >>> print(-coord)
    Coordinate(
        {
           'length': <CartesianPos3D: (x, y, z) [kpc]
               [-1 -2 -3]>
        },
        frame=ICRS()
    )

    """
    return replace(x, data=-x.data)



"""Register primitives."""

__all__: tuple[str, ...] = ()


from dataclasses import replace
from typing import Any, cast

import equinox as eqx
import jax
from jaxtyping import ArrayLike
from quax import register

import quaxed.lax as qlax
import quaxed.numpy as jnp
import unxt as u

from .cartesian import CartesianAcc3D, CartesianPos3D, CartesianVel3D
from .generic import Cartesian3D
from .mathspherical import MathSphericalPos
from coordinax._src.vectors.base_pos import AbstractPos

# ------------------------------------------------


@register(jax.lax.add_p)
def add_cart3d_pos(lhs: CartesianPos3D, rhs: AbstractPos, /) -> CartesianPos3D:
    """Subtract two vectors.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx
    >>> q = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> s = cx.SphericalPos(r=u.Quantity(1, "km"), theta=u.Quantity(90, "deg"),
    ...                     phi=u.Quantity(0, "deg"))
    >>> print(q + s)
    <CartesianPos3D: (x, y, z) [km]
        [2. 2. 3.]>

    """
    cart = rhs.vconvert(CartesianPos3D)
    return jax.tree.map(jnp.add, lhs, cart, is_leaf=u.quantity.is_any_quantity)


@register(jax.lax.add_p)
def add_pp(lhs: CartesianVel3D, rhs: CartesianVel3D, /) -> CartesianVel3D:
    """Add two Cartesian velocities.

    Examples
    --------
    >>> import coordinax as cx

    >>> q = cx.CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> print(q + q)
    <CartesianVel3D: (x, y, z) [km / s]
        [2 4 6]>

    """
    rhs = rhs.uconvert(lhs.units)
    return jax.tree.map(jnp.add, lhs, rhs)


@register(jax.lax.add_p)
def add_aa(lhs: CartesianAcc3D, rhs: CartesianAcc3D, /) -> CartesianAcc3D:
    """Add two Cartesian accelerations.

    Examples
    --------
    >>> import coordinax as cx

    >>> q = cx.vecs.CartesianAcc3D.from_([1, 2, 3], "km/s2")
    >>> print(q + q)
    <CartesianAcc3D: (x, y, z) [km / s2]
        [2 4 6]>

    """
    rhs = rhs.uconvert(lhs.units)
    return jax.tree.map(jnp.add, lhs, rhs)


# ------------------------------------------------
# Dot product
# TODO: see implementation in https://github.com/google/tree-math for how to do
# this more generally.


@register(jax.lax.dot_general_p)
def dot_general_cart3d(
    lhs: CartesianPos3D, rhs: CartesianPos3D, /, **kwargs: Any
) -> u.AbstractQuantity:
    """Dot product of two vectors.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    >>> q1 = cx.vecs.CartesianPos3D.from_([1, 2, 3], "m")
    >>> q2 = cx.vecs.CartesianPos3D.from_([4, 5, 6], "m")

    >>> jnp.dot(q1, q2)
    Quantity(Array(32, dtype=int32), unit='m2')

    """
    return lhs.x * rhs.x + lhs.y * rhs.y + lhs.z * rhs.z


# ------------------------------------------------


@register(jax.lax.mul_p)
def mul_p_vmsph(lhs: ArrayLike, rhs: MathSphericalPos, /) -> MathSphericalPos:
    """Scale the polar position by a scalar.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx
    >>> import quaxed.numpy as jnp

    >>> v = cx.vecs.MathSphericalPos(r=u.Quantity(3, "km"),
    ...                              theta=u.Quantity(90, "deg"),
    ...                              phi=u.Quantity(0, "deg"))

    >>> jnp.linalg.vector_norm(v, axis=-1)
    BareQuantity(Array(3., dtype=float32), unit='km')

    >>> nv = jnp.multiply(2, v)
    >>> print(nv)
    <MathSphericalPos: (r[km], theta[deg], phi[deg])
        [ 6 90  0]>

    """
    # Validation
    lhs = eqx.error_if(
        lhs, any(jax.numpy.shape(lhs)), f"must be a scalar, not {type(lhs)}"
    )
    # Scale the radial distance
    return replace(rhs, r=cast("u.AbstractQuantity", lhs * rhs.r))


@register(jax.lax.mul_p)
def mul_p_arraylike_cart3d(lhs: ArrayLike, rhs: CartesianPos3D, /) -> CartesianPos3D:
    """Scale a position by a scalar.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import coordinax as cx

    >>> v = cx.CartesianPos3D.from_([1, 2, 3], "km")

    >>> print(2 * v)
    <CartesianPos3D: (x, y, z) [km]
        [2 4 6]>

    >>> print(jnp.multiply(2, v))
    <CartesianPos3D: (x, y, z) [km]
        [2 4 6]>

    """
    # Validation
    lhs = eqx.error_if(
        lhs, any(jax.numpy.shape(lhs)), f"must be a scalar, not {type(lhs)}"
    )

    # Scale the components
    return replace(rhs, x=lhs * rhs.x, y=lhs * rhs.y, z=lhs * rhs.z)


# ------------------------------------------------


@register(jax.lax.neg_p)
def neg_p_cart3d_pos(obj: CartesianPos3D, /) -> CartesianPos3D:
    """Negate the `coordinax.CartesianPos3D`.

    Examples
    --------
    >>> import coordinax as cx
    >>> q = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> print(-q)
    <CartesianPos3D: (x, y, z) [km]
        [-1 -2 -3]>

    """
    return jax.tree.map(qlax.neg, obj)


@register(jax.lax.neg_p)
def neg_p_genericcart3d(obj: Cartesian3D, /) -> Cartesian3D:
    """Negate the `coordinax.vecs.Cartesian3D`.

    Examples
    --------
    >>> import coordinax as cx
    >>> q = cx.vecs.Cartesian3D.from_([1, 2, 3], "km")
    >>> print(-q)
    <Cartesian3D: (x, y, z) [km]
    [-1 -2 -3]>

    """
    return jax.tree.map(jnp.negative, obj)


# ------------------------------------------------


@register(jax.lax.sub_p)
def sub_p_cart3d_pos(lhs: CartesianPos3D, rhs: AbstractPos, /) -> CartesianPos3D:
    """Subtract two vectors.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx
    >>> q = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> s = cx.SphericalPos(r=u.Quantity(1, "km"), theta=u.Quantity(90, "deg"),
    ...                     phi=u.Quantity(0, "deg"))
    >>> print(q - s)
    <CartesianPos3D: (x, y, z) [km]
        [0. 2. 3.]>

    """
    cart = rhs.vconvert(CartesianPos3D)
    return jax.tree.map(jnp.subtract, lhs, cart)


@register(jax.lax.sub_p)
def sub_p_v3_v3(lhs: CartesianVel3D, other: CartesianVel3D, /) -> CartesianVel3D:
    """Subtract two differentials.

    Examples
    --------
    >>> from coordinax import CartesianPos3D, CartesianVel3D
    >>> q = CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> print(q - q)
    <CartesianVel3D: (x, y, z) [km / s]
        [0 0 0]>

    """
    return jax.tree.map(jnp.subtract, lhs, other)


@register(jax.lax.sub_p)
def sub_p_a3_a3(lhs: CartesianAcc3D, rhs: CartesianAcc3D, /) -> CartesianAcc3D:
    """Subtract two accelerations.

    Examples
    --------
    >>> import coordinax as cx

    >>> q = cx.vecs.CartesianAcc3D.from_([1, 2, 3], "km/s2")
    >>> print(q - q)
    <CartesianAcc3D: (x, y, z) [km / s2]
        [0 0 0]>

    """
    return jax.tree.map(jnp.subtract, lhs, rhs)



"""Built-in vector classes."""

__all__ = ("SphericalAcc", "SphericalPos", "SphericalVel")

from typing import final

import equinox as eqx

import unxt as u
from dataclassish.converters import Unless

import coordinax._src.custom_types as ct
from .base_spherical import (
    AbstractSphericalAcc,
    AbstractSphericalPos,
    AbstractSphericalVel,
)
from coordinax._src.angles import BatchableAngle
from coordinax._src.distances import AbstractDistance, BatchableDistance, Distance
from coordinax._src.vectors import checks
from coordinax._src.vectors.converters import converter_azimuth_to_range

##############################################################################
# Position


# TODO: make this an alias for SphericalPolarPos, the more correct description?
@final
class SphericalPos(AbstractSphericalPos):
    """Spherical-Polar coordinates.

    .. note::

        This class follows the Physics conventions (ISO 80000-2:2019).

    Parameters
    ----------
    r
        Radial distance r (slant distance to origin),
    theta
        Polar angle [0, 180] [deg] where 0 is the z-axis.
    phi
        Azimuthal angle [0, 360) [deg] where 0 is the x-axis.

    """

    r: BatchableDistance = eqx.field(converter=Unless(AbstractDistance, Distance.from_))
    r"""Radial distance :math:`r \in [0,+\infty)`."""

    theta: BatchableAngle = eqx.field(converter=u.Angle.from_)
    r"""Inclination angle :math:`\theta \in [0,180]`."""

    phi: BatchableAngle = eqx.field(
        converter=Unless(
            u.Angle, lambda x: converter_azimuth_to_range(u.Angle.from_(x))
        )
    )
    r"""Azimuthal angle, generally :math:`\phi \in [0,360)`."""

    def __check_init__(self) -> None:
        """Check the validity of the initialization."""
        checks.check_polar_range(self.theta)


@final
class SphericalVel(AbstractSphericalVel):
    """Spherical velocity."""

    r: ct.BBtSpeed = eqx.field(converter=u.Quantity["speed"].from_)
    r"""Radial speed :math:`dr/dt \in [-\infty, \infty]."""

    theta: ct.BBtAngularSpeed = eqx.field(converter=u.Quantity["angular speed"].from_)
    r"""Inclination speed :math:`d\theta/dt \in [-\infty, \infty]."""

    phi: ct.BBtAngularSpeed = eqx.field(converter=u.Quantity["angular speed"].from_)
    r"""Azimuthal speed :math:`d\phi/dt \in [-\infty, \infty]."""


@final
class SphericalAcc(AbstractSphericalAcc):
    """Spherical differential representation."""

    r: ct.BBtAcc = eqx.field(converter=u.Quantity["acceleration"].from_)
    r"""Radial acceleration :math:`d^2r/dt^2 \in [-\infty, \infty]."""

    theta: ct.BBtAngularAcc = eqx.field(
        converter=u.Quantity["angular acceleration"].from_
    )
    r"""Inclination acceleration :math:`d^2\theta/dt^2 \in [-\infty, \infty]."""

    phi: ct.BBtAngularAcc = eqx.field(
        converter=u.Quantity["angular acceleration"].from_
    )
    r"""Azimuthal acceleration :math:`d^2\phi/dt^2 \in [-\infty, \infty]."""



"""Built-in vector classes."""

__all__ = (
    "ProlateSpheroidalAcc",
    "ProlateSpheroidalPos",
    "ProlateSpheroidalVel",
)

from dataclasses import KW_ONLY
from typing import final

import equinox as eqx
from jaxtyping import Shaped

import quaxed.numpy as jnp
import unxt as u
from dataclassish.converters import Unless

import coordinax._src.custom_types as ct
from .base import AbstractAcc3D, AbstractPos3D, AbstractVel3D
from coordinax._src.angles import BatchableAngleQ
from coordinax._src.vectors import checks
from coordinax._src.vectors.base import VectorAttribute
from coordinax._src.vectors.converters import converter_azimuth_to_range


@final
class ProlateSpheroidalPos(AbstractPos3D):
    r"""Prolate spheroidal coordinates as defined by Dejonghe & de Zeeuw 1988.

    Note that valid coordinates have: $- mu >= \Delta^2 - |nu| <= \Delta^2 -
    \Delta > 0$

    Parameters
    ----------
    mu
        The spheroidal mu coordinate. This is called `lambda` by Dejonghe & de
        Zeeuw. Surfaces of constant mu are ellipsoids with foci at the origin
        and at (`Delta`, 0, 0).
    nu
        The spheroidal nu coordinate. Surfaces of constant nu are hyperboloids
        of two sheets.
    phi
        Azimuthal angle [0, 360) [deg] where 0 is the x-axis.
    Delta
        The focal length of the coordinate system. Must be > 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> vec = cxv.ProlateSpheroidalPos(
    ...     mu=u.Quantity(3.0, "km2"),
    ...     nu=u.Quantity(0.5, "km2"),
    ...     phi=u.Quantity(0.25, "rad"),
    ...     Delta=u.Quantity(1.5, "km"),
    ... )
    >>> print(vec)
    <ProlateSpheroidalPos: (mu[km2], nu[km2], phi[rad])
     Delta=Quantity(1.5, unit='km')
        [3.   0.5  0.25]>

    This fails with a zero or negative Delta:

    >>> try: vec = cxv.ProlateSpheroidalPos(
    ...     mu=u.Quantity(3.0, "km2"),
    ...     nu=u.Quantity(0.5, "km2"),
    ...     phi=u.Quantity(0.25, "rad"),
    ...     Delta=u.Quantity(0.0, "km"),
    ... )
    ... except Exception as e: pass

    Or with invalid mu and nu:

    >>> try: vec = cxv.ProlateSpheroidalPos(
    ...     mu=u.Quantity(0.5, "km2"),
    ...     nu=u.Quantity(0.5, "km2"),
    ...     phi=u.Quantity(0.25, "rad"),
    ...     Delta=u.Quantity(1.5, "km"),
    ... )
    ... except Exception as e: pass

    We can convert to other coordinates:

    >>> sph = vec.vconvert(cxv.SphericalPos)
    >>> print(sph)
    <SphericalPos: (r[km], theta[rad], phi[rad])
        [1.118 0.752 0.25 ]>

    However, this is generally a one-way conversion, as the focal length
    parameter `Delta` is not retained through the conversion. To convert back to
    prolate spheroidal coordinates, we need to provide the focal length again:

    >>> vec2 = sph.vconvert(cxv.ProlateSpheroidalPos, Delta=u.Quantity(1.5, "km"))
    >>> print(vec2.round(3))
    <ProlateSpheroidalPos: (mu[km2], nu[km2], phi[rad])
     Delta=Quantity(1.5, unit='km')
        [3.   0.5  0.25]>

    >>> print((vec2 - vec).vconvert(cxv.CartesianPos3D))
    <CartesianPos3D: (x, y, z) [km]
        [0. 0. 0.]>

    """

    mu: ct.BBtArea = eqx.field(converter=u.Quantity["area"].from_)
    r"""Spheroidal mu coordinate :math:`\mu \in [0,+\infty)` (called :math:`\lambda` in
     some Galactic dynamics contexts)."""

    nu: ct.BBtArea = eqx.field(converter=u.Quantity["area"].from_)
    r"""Spheroidal nu coordinate :math:`\lambda \in [-\infty,+\infty)`."""

    phi: BatchableAngleQ = eqx.field(
        converter=Unless(
            u.Angle, lambda x: converter_azimuth_to_range(u.Angle.from_(x))
        )
    )
    r"""Azimuthal angle, generally :math:`\phi \in [0,360)`."""

    _: KW_ONLY
    Delta: Shaped[u.Quantity["length"], ""] = VectorAttribute()
    """Focal length of the coordinate system."""

    def __check_init__(self) -> None:
        """Check the validity of the initialization."""
        checks.check_non_negative_non_zero(self.Delta, name="Delta")
        checks.check_greater_than_equal(
            self.mu, self.Delta**2, name="mu", comparison_name="Delta^2"
        )
        checks.check_less_than_equal(
            jnp.abs(self.nu), self.Delta**2, name="nu", comparison_name="Delta^2"
        )


@final
class ProlateSpheroidalVel(AbstractVel3D):
    r"""Prolate spheroidal differential representation.

    Parameters
    ----------
    mu
        Prolate spheroidal mu speed $d\mu/dt \in [-\infty, \infty]$.
    nu
        Prolate spheroidal nu speed $d\nu/dt \in [-\infty, \infty]$.
    phi
        Azimuthal speed $d\phi/dt \in [-\infty, \infty]$.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> x = cxv.CartesianPos3D.from_(u.Quantity([1, 2, 3], "kpc"))
    >>> v = cxv.CartesianVel3D.from_(u.Quantity([4, 5, 6], "km/s"))

    >>> px = x.vconvert(cxv.ProlateSpheroidalPos, Delta=u.Quantity(4, "kpc"))
    >>> pv = v.vconvert(cxv.ProlateSpheroidalVel, px)

    >>> print(pv.vconvert(cxv.CartesianVel3D, px))
    <CartesianVel3D: (x, y, z) [km / s]
        [4. 5. 6.]>

    >>> print(pv.vconvert(cxv.CartesianVel3D, x, Delta=u.Quantity(4, "kpc")))
    <CartesianVel3D: (x, y, z) [km / s]
        [4. 5. 6.]>

    """

    mu: ct.BBtKinematicFlux = eqx.field(converter=u.Quantity["diffusivity"].from_)
    r"""Prolate spheroidal mu speed $d\mu/dt \in [-\infty, \infty]$."""

    nu: ct.BBtKinematicFlux = eqx.field(converter=u.Quantity["diffusivity"].from_)
    r"""Prolate spheroidal nu speed $d\nu/dt \in [-\infty, \infty]$."""

    phi: ct.BBtAngularSpeed = eqx.field(converter=u.Quantity["angular speed"].from_)
    r"""Azimuthal speed $d\phi/dt \in [-\infty, \infty]$."""


@final
class ProlateSpheroidalAcc(AbstractAcc3D):
    r"""Prolate spheroidal acceleration representation.

    Parameters
    ----------
    mu
        Prolate spheroidal mu speed $d\mu/dt \in [-\infty, \infty]$.
    nu
        Prolate spheroidal nu speed $d\nu/dt \in [-\infty, \infty]$.
    phi
        Azimuthal speed $d\phi/dt \in [-\infty, \infty]$.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> x = cxv.CartesianPos3D.from_(u.Quantity([1, 2, 3], "kpc"))
    >>> v = cxv.CartesianVel3D.from_(u.Quantity([4, 5, 6], "km/s"))
    >>> a = cxv.CartesianAcc3D.from_(u.Quantity([4, 5, 6], "km/s2"))

    >>> px = x.vconvert(cxv.ProlateSpheroidalPos, Delta=u.Quantity(4, "kpc"))
    >>> pa = a.vconvert(cxv.ProlateSpheroidalAcc, v, px)

    >>> print(pa.vconvert(cxv.CartesianAcc3D, v, px))
    <CartesianAcc3D: (x, y, z) [km / s2]
        [4. 5. 6.]>

    >>> print(pa.vconvert(cxv.CartesianAcc3D, v, x, Delta=u.Quantity(4, "kpc")))
    <CartesianAcc3D: (x, y, z) [km / s2]
        [4. 5. 6.]>

    """

    mu: ct.BBtSpecificEnergy = eqx.field(converter=u.Quantity["specific energy"].from_)
    r"""Prolate spheroidal mu acceleration $d^2\mu/dt^2 \in [-\infty, \infty]$."""

    nu: ct.BBtSpecificEnergy = eqx.field(converter=u.Quantity["specific energy"].from_)
    r"""Prolate spheroidal nu acceleration $d^2\nu/dt^2 \in [-\infty, \infty]$."""

    phi: ct.BBtAngularAcc = eqx.field(
        converter=u.Quantity["angular acceleration"].from_
    )
    r"""Azimuthal acceleration $d^2\phi/dt^2 \in [-\infty, \infty]$."""



"""Distance constructors."""

__all__: tuple[str, ...] = ()

from typing import Any

import unxt as u

from .base import AbstractDistance
from .funcs import distance, distance_modulus, parallax
from .measures import Distance, DistanceModulus, Parallax

parallax_base_length = u.Quantity(1, "AU")
distance_modulus_base_distance = u.Quantity(10, "pc")


@u.AbstractQuantity.from_.dispatch
def from_(
    cls: type[Distance],
    obj: AbstractDistance
    | u.Quantity["length"]
    | u.Quantity["angle"]
    | u.Quantity["mag"],
    /,
    **kw: Any,
) -> Distance:
    """Construct a `Distance` from the inputs."""
    return distance(obj, **kw)


@u.AbstractQuantity.from_.dispatch
def from_(
    cls: type[DistanceModulus],
    dm: AbstractDistance
    | u.Quantity["mag"]
    | u.Quantity["length"]
    | u.Quantity["angle"],
    /,
    **kwargs: Any,
) -> DistanceModulus:
    """Construct a `DistanceModulus` from the input."""
    return distance_modulus(dm, **kwargs)


@u.AbstractQuantity.from_.dispatch
def from_(
    cls: type[Parallax],
    obj: AbstractDistance
    | u.Quantity["angle"]
    | u.Quantity["length"]
    | u.Quantity["mag"],
    /,
    **kwargs: Any,
) -> Parallax:
    """Construct a `Parallax` the input."""
    return parallax(obj, **kwargs)



"""Galilean coordinate transformations."""

__all__ = ["GalileanTranslation"]


from typing import Any, Literal, cast, final

import equinox as eqx
from plum import convert, dispatch

import quaxed.numpy as jnp
import unxt as u
from dataclassish.converters import Unless

from .base import AbstractGalileanOperator
from coordinax._src.operators.base import AbstractOperator
from coordinax._src.operators.identity import Identity
from coordinax._src.vectors import api
from coordinax._src.vectors.base_pos import AbstractPos
from coordinax._src.vectors.d3 import AbstractPos3D, CartesianPos3D


@final
class GalileanTranslation(AbstractGalileanOperator):
    r"""Operator for spatio-temporal translations.

    The coordinate transform is given by:

    .. math::

        (t,\mathbf{x}) \mapsto (t+s, \mathbf{x} + \mathbf {a})

    where :math:`a \in R^3` and :math:`s \in R`.  Therefore for a potential
    :math:`\Phi(t,\mathbf{x})` in the translated frame the potential is given by
    the subtraction of the translation.

    Parameters
    ----------
    delta_t
        The time translation.
    delta_q
        The spatial translation vector.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    We can then create a translation operator:

    >>> op = cx.ops.GalileanTranslation.from_([1.0, 2.0, 3.0, 4.0], "km")
    >>> op
    GalileanTranslation(
      delta_t=Quantity(f32[], unit='s'),
      delta_q=CartesianPos3D( ... )
    )

    Note that the translation is a `unxt.Quantity` and a
    `coordinax.vecs.AbstractPos`, which was constructed from a 1D array, using
    :meth:`coordinax.vecs.AbstractPos.from_`.  We can also construct it
    directly, which allows for other vector types.

    >>> qshift = cx.SphericalPos(r=u.Quantity(1.0, "km"),
    ...                          theta=u.Quantity(jnp.pi/2, "rad"),
    ...                          phi=u.Quantity(0, "rad"))
    >>> op = cx.ops.GalileanTranslation(u.Quantity(1.0, "Gyr"), qshift)
    >>> op
    GalileanTranslation(
      delta_t=Quantity(weak_f32[], unit='Gyr'),
      delta_q=SphericalPos( ... )
    )

    Translation operators can be applied to `coordinax.vecs.AbstractPos` and
    `unxt.Quantity`:

    >>> q = cx.CartesianPos3D.from_([0, 0, 0], "km")
    >>> t = u.Quantity(0, "Gyr")
    >>> newt, newq = op(t, q)
    >>> newq.x
    Quantity(Array(1., dtype=float32, ...), unit='km')
    >>> newt
    Quantity(Array(1., dtype=float32, ...), unit='Gyr')

    """

    delta_t: u.Quantity["time"] = eqx.field(converter=u.Quantity.from_)
    """The time translation.

    This parameter uses :meth:`unxt.Quantity.from_` to enable a variety of more
    convenient input types. See `unxt.Quantity` for details.
    """

    delta_q: AbstractPos = eqx.field(converter=Unless(AbstractPos, api.vector))
    """The spatial translation.

    This parameters uses :meth:`coordinax.vecs.CartesianPos3D.from_` to enable a
    variety of more convenient input types. See `coordinax.vecs.CartesianPos3D`
    for details.
    """

    # -------------------------------------------

    @property
    def is_inertial(self) -> Literal[True]:
        """Galilean translation is an inertial-frame preserving transformation.

        Examples
        --------
        >>> import coordinax as cx

        >>> op = cx.ops.GalileanTranslation.from_([0, 1, 1, 1], "km")

        >>> op.is_inertial
        True

        """
        return True

    @property
    def inverse(self) -> "GalileanTranslation":
        """The inverse of the operator.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> qshift = cx.CartesianPos3D.from_([1, 1, 1], "km")
        >>> op = cx.ops.GalileanTranslation(u.Quantity(1, "Gyr"), qshift)

        >>> print(op.inverse)
        GalileanTranslation(
            delta_t=Quantity(-1, unit='Gyr'),
            delta_q=<CartesianPos3D: (x, y, z) [km]
                [-1 -1 -1]>
        )

        """
        delta_q = cast("AbstractPos3D", -self.delta_q)
        return GalileanTranslation(-self.delta_t, delta_q)

    # -------------------------------------------

    @AbstractOperator.__call__.dispatch  # type: ignore[untyped-decorator]
    def __call__(
        self: "GalileanTranslation",
        t: u.Quantity["time"],
        x: AbstractPos3D,
        /,
        **__: Any,
    ) -> tuple[u.Quantity["time"], AbstractPos3D]:
        """Apply the translation to the coordinates.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx
        >>> import coordinax.ops as cxo

        Explicitly construct the translation operator:

        >>> qshift = cx.CartesianPos3D.from_([1, 1, 1], "km")
        >>> tshift = u.Quantity(1, "Gyr")
        >>> op = cx.ops.GalileanTranslation(tshift, qshift)

        Construct a vector to translate

        >>> q = cx.CartesianPos3D.from_([1, 2, 3], "km")
        >>> t = u.Quantity(1, "Gyr")
        >>> newt, newq = op(t, q)

        >>> newq.x
        Quantity(Array(2, dtype=int32), unit='km')

        >>> newt
        Quantity(Array(2, dtype=int32, ...), unit='Gyr')

        """
        return t + self.delta_t, x + self.delta_q


@AbstractOperator.from_.dispatch  # type: ignore[untyped-decorator]
def from_(
    cls: type[GalileanTranslation], delta: u.AbstractQuantity, /
) -> GalileanTranslation:
    """Construct from a time+space array.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.ops as cxo

    >>> q = u.Quantity([1.0, 2.0, 3.0, 4.0], "km")
    >>> op = cxo.GalileanTranslation.from_(q)

    """
    return cls(
        delta_t=delta[0]
        / u.Quantity(299_792.458, "km/s"),  # TODO: couple to FourVector value
        delta_q=CartesianPos3D.from_(delta[1:]),
    )


# -------------------------------------------


@dispatch
def simplify_op(
    op: GalileanTranslation, /, **kwargs: Any
) -> GalileanTranslation | Identity:
    """Simplify a Galilean translation operator.

    Examples
    --------
    >>> import coordinax.ops as cxo

    An operator with real effect cannot be simplified:

    >>> op = cxo.GalileanTranslation.from_([3e8, 1, 0, 0], "m")
    >>> cxo.simplify_op(op)
    GalileanTranslation(
      delta_t=Quantity(f32[], unit='m s / km'),
      delta_q=CartesianPos3D( ... )
    )

    An operator with no effect can be simplified:

    >>> op = cxo.GalileanTranslation.from_([0, 0, 0, 0], "m")
    >>> cxo.simplify_op(op)
    Identity()

    """
    # Check if the translation is zero.
    q = convert(op.delta_q, u.Quantity).value
    if jnp.allclose(op.delta_t.value, 0, **kwargs) and jnp.allclose(q, 0, **kwargs):
        return Identity()
    # TODO: Check if the translation is purely spatial.

    return op


# TODO: show op3.translation = op1.translation + op2.translation
@dispatch
def simplify_op(
    op1: GalileanTranslation, op2: GalileanTranslation, /
) -> GalileanTranslation:
    """Combine two translations into a single translation.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> qshift = cx.CartesianPos3D.from_([1, 0, 0], "km")
    >>> tshift = u.Quantity(1, "Gyr")
    >>> op1 = cx.ops.GalileanTranslation(tshift, qshift)

    >>> qshift = cx.CartesianPos3D.from_([0, 1, 0], "km")
    >>> op2 = cx.ops.GalileanTranslation(tshift, qshift)

    >>> op3 = cx.ops.simplify_op(op1, op2)
    >>> op3
    GalileanTranslation(
      delta_t=Quantity(weak_i32[], unit='Gyr'),
      delta_q=CartesianPos3D( ... )
    )

    """
    return GalileanTranslation(op1.delta_t + op2.delta_t, op1.delta_q + op2.delta_q)



"""Representation of coordinates in different systems."""

__all__ = ("KinematicSpace",)

import math
from collections.abc import Callable, ItemsView, Iterable, KeysView, Mapping, ValuesView
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, Generic, cast, final
from typing_extensions import override

import equinox as eqx
import jax
import wadler_lindig as wl
from astropy.units import PhysicalType as Dimension
from jax import Device
from plum import dispatch

import quaxed.numpy as jnp
import unxt as u
from dataclassish import replace
from xmmutablemap import ImmutableMap

from .utils import DimensionLike, _get_dimension_name, can_broadcast_shapes
from coordinax._src.custom_types import Unit
from coordinax._src.utils import classproperty
from coordinax._src.vectors.api import vector
from coordinax._src.vectors.base import AbstractVector, AbstractVectorLike
from coordinax._src.vectors.base_pos import PosT

if TYPE_CHECKING:
    from typing import Self


# TODO: figure out how to make the keys into Dimension objects, not str. This is
#       running afoul of Jax's tree flattening, where ImmutableMap and
#       eqx.Module differ.
@final
class KinematicSpace(
    AbstractVectorLike,
    ImmutableMap[str, AbstractVector],  # type: ignore[misc]
    Generic[PosT],
):
    """A collection of vectors that acts like the primary vector.

    Parameters
    ----------
    *args, **kwargs
        See input to `dict` for the input data.

    Examples
    --------
    >>> import coordinax as cx

    >>> x = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> v = cx.CartesianVel3D.from_([4, 5, 6], "km/s")
    >>> a = cx.vecs.CartesianAcc3D.from_([7, 8, 9], "km/s^2")

    All the vectors can be brought together into a space:

    >>> space = cx.KinematicSpace(length=x, speed=v, acceleration=a)
    >>> print(space)
    KinematicSpace({
       'length': <CartesianPos3D: (x, y, z) [km]
           [1 2 3]>,
       'speed': <CartesianVel3D: (x, y, z) [km / s]
           [4 5 6]>,
       'acceleration': <CartesianAcc3D: (x, y, z) [km / s2]
           [7 8 9]>
    })

    The vectors can initialized from `unxt.Quantity` objects and can have
    (brodcastable) batch shapes:

    >>> w = cx.KinematicSpace(
    ...     length=u.Quantity([[8.5, 0, 0], [10, 0, 0]], "kpc"),
    ...     speed=u.Quantity([0, 200, 0], "km/s"))
    >>> print(w)
    KinematicSpace({
       'length': <CartesianPos3D: (x, y, z) [kpc]
           [[ 8.5  0.   0. ]
            [10.   0.   0. ]]>,
       'speed': <CartesianVel3D: (x, y, z) [km / s]
           [  0 200   0]>
    })

    The vectors can then be accessed by key:

    >>> space["length"]
    CartesianPos3D( ... )

    The space can also be converted to different representations. If the
    conversion is on the primary vector, the other vectors will be
    correspondingly converted.

    >>> space.vconvert(cx.SphericalPos)
    KinematicSpace({
        'length': SphericalPos( ... ),
        'speed': SphericalVel( ... ),
        'acceleration': SphericalAcc( ... )
    })

    >>> cx.vconvert(cx.SphericalPos, space)
    KinematicSpace({
        'length': SphericalPos( ... ),
        'speed': SphericalVel( ... ),
        'acceleration': SphericalAcc( ... )
    })

    Actions on the space are done on the contained vectors.

    >>> w = cx.KinematicSpace(
    ...     length=cx.CartesianPos3D.from_([[[1, 2, 3], [4, 5, 6]]], "m"),
    ...     speed=cx.CartesianVel3D.from_([[[1, 2, 3], [4, 5, 6]]], "m/s")
    ... )

    >>> w.ndim
    2

    >>> w.shape
    (1, 2)

    >>> w.shapes
    mappingproxy({'length': (1, 2), 'speed': (1, 2)})

    >>> w.mT.shapes
    mappingproxy({'length': (2, 1), 'speed': (2, 1)})

    There are convenience ways to initialize the vectors in the space:

    >>> space = cx.KinematicSpace.from_({"length": u.Quantity([1, 2, 3], "km"),
    ...                         "speed": u.Quantity([4, 5, 6], "km/s")})
    >>> print(space)
    KinematicSpace({
       'length': <CartesianPos3D: (x, y, z) [km]
           [1 2 3]>,
       'speed': <CartesianVel3D: (x, y, z) [km / s]
           [4 5 6]>
    })

    """

    # TODO: https://peps.python.org/pep-0728/#the-extra-items-class-parameter
    _data: dict[str, AbstractVector] = eqx.field(repr=False)

    def __init__(  # pylint: disable=super-init-not-called  # TODO: resolve this
        self,
        /,
        *args: Mapping[DimensionLike, Any]
        | tuple[DimensionLike, Any]
        | Iterable[tuple[DimensionLike, Any]],
        **kwargs: Any,
    ) -> None:
        # Fast-path for the common case of a single dict input.
        if len(args) == 1 and isinstance(args[0], Mapping) and not kwargs:
            self._data = dict(args[0])
            return

        # Consolidate the inputs into a single dict, then process keys & values.
        raw = dict(*args, **kwargs)  # process the input data

        # Process the keys
        dims = [u.dimension(k) for k in raw]
        keys = [_get_dimension_name(dim) for dim in dims]
        # Convert the values to vectors
        values = [vector(v) for v in raw.values()]

        # Check that the shapes are broadcastable
        values = eqx.error_if(
            values,
            not can_broadcast_shapes([v.shape for v in values]),
            "vector shapes are not broadcastable.",
        )

        self._data = dict(zip(keys, values, strict=True))

    def __check_init__(self) -> None:
        if "length" not in self._data:
            msg = "KinematicSpace must have a 'length' vector component."
            raise ValueError(msg)

    @classmethod
    def _dimensionality(cls) -> int:
        """Dimensionality of the Space.

        Examples
        --------
        >>> import coordinax as cx

        >>> w = cx.KinematicSpace.from_(cx.CartesianPos3D.from_([1, 2, 3], "kpc"))
        >>> try: w._dimensionality()
        ... except NotImplementedError as e: print("not implemented")
        not implemented

        """
        raise NotImplementedError  # TODO: implement this

    @property
    def q(self) -> PosT:
        """Get the position vector of the space."""
        return cast("PosT", self._data["length"])

    # ===============================================================
    # Mapping API

    @dispatch
    def __getitem__(self, key: Any) -> Any:
        """Get the vector by key.

        Examples
        --------
        >>> import coordinax as cxv
        >>> w = cxv.KinematicSpace(
        ...     length=cxv.CartesianPos3D.from_([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cxv.CartesianVel3D.from_([[[1, 2, 3], [4, 5, 6]]], "m/s"))

        By number:

        >>> w[0]
        KinematicSpace({
            'length': CartesianPos3D(
                x=Quantity([1, 4], unit='m'),
                y=Quantity([2, 5], unit='m'),
                z=Quantity([3, 6], unit='m')
            ),
            'speed': CartesianVel3D(
                x=Quantity([1, 4], unit='m / s'),
                y=Quantity([2, 5], unit='m / s'),
                z=Quantity([3, 6], unit='m / s')
            )
        })

        By slice:

        >>> w[1:]
        KinematicSpace({
            'length': CartesianPos3D(
                x=Quantity([], unit='m'), y=Quantity([], unit='m'),
                z=Quantity([], unit='m')
            ),
            'speed': CartesianVel3D(
                x=Quantity([], unit='m / s'),
                y=Quantity([], unit='m / s'),
                z=Quantity([], unit='m / s')
            )
        })

        By Ellipsis:

        >>> w[...]
        KinematicSpace({
            'length': CartesianPos3D(
                x=Quantity([[1, 4]], unit='m'),
                y=Quantity([[2, 5]], unit='m'),
                z=Quantity([[3, 6]], unit='m')
            ),
            'speed': CartesianVel3D(
                x=Quantity([[1, 4]], unit='m / s'),
                y=Quantity([[2, 5]], unit='m / s'),
                z=Quantity([[3, 6]], unit='m / s')
            )
        })

        By tuple[int, ...]:

        >>> w[(0, 1)]
        KinematicSpace({
            'length': CartesianPos3D(
                x=Quantity(4, unit='m'), y=Quantity(5, unit='m'),
                z=Quantity(6, unit='m')
            ),
            'speed': CartesianVel3D(
                x=Quantity(4, unit='m / s'),
                y=Quantity(5, unit='m / s'),
                z=Quantity(6, unit='m / s')
            )
        })

        This also supports numpy index arrays. But this example section
        highlights core python indexing.

        """
        return KinematicSpace(**{k: v[key] for k, v in self.items()})

    @dispatch
    def __getitem__(self, key: str | Dimension) -> Any:
        """Get the vector by key.

        Examples
        --------
        >>> import coordinax as cx
        >>> w = cx.KinematicSpace(
        ...     length=cx.CartesianPos3D.from_([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVel3D.from_([[[1, 2, 3], [4, 5, 6]]], "m/s"))

        By string key (which is the dimension name):

        >>> print(w["length"])
        <CartesianPos3D: (x, y, z) [m]
            [[[1 2 3]
              [4 5 6]]]>

        By the actual dimension object:

        >>> import unxt as u
        >>> print(w[u.dimension("length")])
        <CartesianPos3D: (x, y, z) [m]
            [[[1 2 3]
              [4 5 6]]]>

        """
        if isinstance(key, Dimension):
            key = _get_dimension_name(key)

        return ImmutableMap.__getitem__(self, key)

    def keys(self) -> KeysView[str]:
        return self._data.keys()

    def values(self) -> ValuesView[AbstractVector]:
        return self._data.values()

    def items(self) -> ItemsView[str, AbstractVector]:
        return self._data.items()

    # ===============================================================
    # Quax API

    # TODO: unify this with AvalMixin
    def aval(self) -> jax.core.ShapedArray:
        """Return the vector as a JAX array.

        Examples
        --------
        >>> import coordinax as cx
        >>> w = cx.KinematicSpace(
        ...     length=cx.CartesianPos3D.from_([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVel3D.from_([7, 8, 9], "m/s")
        ... )
        >>> w.aval()
        ShapedArray(int32[1,2,6])

        """
        avals = [v.aval() for v in self.values()]
        shapes = [a.shape for a in avals]
        shape = (
            *jnp.broadcast_shapes(*[s[:-1] for s in shapes]),
            sum(s[-1] for s in shapes),
        )
        dtype = jnp.result_type(*map(jnp.dtype, avals))
        return jax.core.ShapedArray(shape, dtype)

    @override
    @property
    def shape(self) -> tuple[int, ...]:
        """Get the shape of the vector's components.

        When represented as a single array, the vector has an additional
        dimension at the end for the components.

        """
        return jnp.broadcast_shapes(*[v.shape for v in self.values()])

    # ===============================================================
    # Array API

    # ---------------------------------------------------------------
    # Attributes

    @override
    @property
    def ndim(self) -> int:
        """Get the number of dimensions of the vector.

        When represented as a single array, the vector has an additional
        dimension at the end for the components.

        Examples
        --------
        >>> import coordinax as cx

        >>> w = cx.KinematicSpace(
        ...     length=cx.CartesianPos3D.from_([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVel3D.from_([7, 8, 9], "m/s")
        ... )

        >>> w.ndim
        2

        """
        return len(self.shape)

    @property
    def size(self) -> int:
        """Total number of elements in the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> w = cx.KinematicSpace(
        ...     length=cx.CartesianPos3D.from_([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVel3D.from_([[[1, 2, 3], [4, 5, 6]]], "m/s") )

        >>> w.size
        2

        """
        return math.prod(self.shape)

    @property
    def mT(self) -> "Self":  # noqa: N802
        """Transpose each vector in the space."""
        return replace(self, **{k: v.mT for k, v in self.items()})

    @property
    def T(self) -> "Self":  # noqa: N802
        """Transpose each vector in the space.

        Examples
        --------
        >>> import coordinax as cx

        >>> w = cx.KinematicSpace(
        ...     length=cx.CartesianPos3D.from_([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVel3D.from_([[[1, 2, 3], [4, 5, 6]]], "m/s")
        ... )

        >>> w.T.shapes
        mappingproxy({'length': (2, 1), 'speed': (2, 1)})

        """
        return replace(self, **{k: v.T for k, v in self.items()})

    # ---------------------------------------------------------------
    # Methods

    def __pdoc__(self, **kwargs: Any) -> wl.AbstractDoc:
        """Return the Wadler-Lindig representation.

        Examples
        --------
        >>> import coordinax as cx

        >>> q = cx.CartesianPos3D.from_([1, 2, 3], "m")
        >>> p = cx.CartesianVel3D.from_([1, 2, 3], "m/s")
        >>> w = cx.KinematicSpace(length=q, speed=p)
        >>> w
        KinematicSpace({
            'length': CartesianPos3D(
                x=Quantity(1, unit='m'),
                y=Quantity(2, unit='m'),
                z=Quantity(3, unit='m')
            ),
            'speed': CartesianVel3D(
                x=Quantity(1, unit='m / s'),
                y=Quantity(2, unit='m / s'),
                z=Quantity(3, unit='m / s')
            )
        })

        """
        data_pdoc = wl.pdoc(self._data, **kwargs)
        data_pdoc = replace(  # remove the "{}""
            data_pdoc, {"child": {"children": data_pdoc.child.children[1:-1]}}
        )
        return (
            wl.TextDoc(self.__class__.__name__)
            + wl.TextDoc("({")
            + data_pdoc
            + wl.TextDoc("})")
        )

    # ===============================================================
    # Python API

    def __str__(self) -> str:  # TODO: update using wadler-lindig
        """Return the string representation.

        Examples
        --------
        >>> import coordinax as cx

        >>> q = cx.CartesianPos3D.from_([1, 2, 3], "m")
        >>> p = cx.CartesianVel3D.from_([4, 5, 6], "m/s")
        >>> w = cx.KinematicSpace(length=q, speed=p)
        >>> print(w)
        KinematicSpace({
            'length': <CartesianPos3D: (x, y, z) [m]
                [1 2 3]>,
            'speed': <CartesianVel3D: (x, y, z) [m / s]
                [4 5 6]>
        })

        """
        return wl.pformat(self, vector_form=True, short_arrays=False)

    # ===============================================================

    def asdict(
        self,
        *,
        dict_factory: Callable[[Any], Mapping[str, AbstractVector]] = dict,
    ) -> Mapping[str, AbstractVector]:
        """Return the vector collection as a Mapping.

        See Also
        --------
        `dataclasses.asdict`
            This applies recursively to the components of the vector.

        """
        return dict_factory(self._data)

    @classproperty
    @classmethod
    def components(cls) -> tuple[str, ...]:
        """Vector component names."""
        raise NotImplementedError  # TODO: implement this

    @property
    def units(self) -> MappingProxyType[str, Unit]:
        """Get the units of the vector's components."""
        raise NotImplementedError  # TODO: implement this

    @property
    def dtypes(self) -> MappingProxyType[str, MappingProxyType[str, jnp.dtype[Any]]]:
        """Get the dtypes of the vector's components.

        Examples
        --------
        >>> import coordinax as cx

        >>> w = cx.KinematicSpace(
        ...     length=cx.CartesianPos3D.from_([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVel3D.from_([[[1, 2, 3], [4, 5, 6]]], "m/s")
        ... )

        >>> w.dtypes
        mappingproxy({'length': mappingproxy({'x': dtype('int32'), 'y': dtype('int32'), 'z': dtype('int32')}),
                      'speed': mappingproxy({'x': dtype('int32'), 'y': dtype('int32'), 'z': dtype('int32')})})

        """  # noqa: E501
        return MappingProxyType({k: v.dtypes for k, v in self._data.items()})

    @property
    def devices(self) -> MappingProxyType[str, MappingProxyType[str, Device]]:
        """Get the devices of the vector's components.

        Examples
        --------
        >>> import coordinax as cx

        >>> w = cx.KinematicSpace(
        ...     length=cx.CartesianPos3D.from_([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVel3D.from_([[[1, 2, 3], [4, 5, 6]]], "m/s")
        ... )

        >>> w.devices
        mappingproxy({'length': mappingproxy({'x': CpuDevice(id=0), 'y': CpuDevice(id=0), 'z': CpuDevice(id=0)}),
                      'speed': mappingproxy({'x': CpuDevice(id=0), 'y': CpuDevice(id=0), 'z': CpuDevice(id=0)})})

        """  # noqa: E501
        return MappingProxyType({k: v.devices for k, v in self._data.items()})

    @property
    def shapes(self) -> MappingProxyType[str, tuple[int, ...]]:
        """Get the shapes of the spaces's fields.

        Examples
        --------
        >>> import coordinax as cx

        >>> w = cx.KinematicSpace(
        ...     length=cx.CartesianPos3D.from_([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVel3D.from_([[[1, 2, 3], [4, 5, 6]]], "m/s")
        ... )

        >>> w.shapes
        mappingproxy({'length': (1, 2), 'speed': (1, 2)})

        """
        return MappingProxyType({k: v.shape for k, v in self._data.items()})

    @property
    def sizes(self) -> MappingProxyType[str, int]:
        """Get the sizes of the vector's components.

        Examples
        --------
        >>> import coordinax as cx

        >>> w = cx.KinematicSpace(
        ...     length=cx.CartesianPos3D.from_([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVel3D.from_([[[1, 2, 3], [4, 5, 6]]], "m/s")
        ... )

        >>> w.sizes
        mappingproxy({'length': 6, 'speed': 6})

        """
        return MappingProxyType({k: v.size for k, v in self._data.items()})



"""Utilities for Space. Private module."""

__all__: tuple[str, ...] = ()

from collections.abc import Sequence
from typing import TypeAlias

from astropy.units import PhysicalType as Dimension

import quaxed.numpy as jnp
import unxt as u
from unxt._src.dimensions import name_of

DimensionLike: TypeAlias = Dimension | str
Shape: TypeAlias = tuple[int, ...]


def _get_dimension_name(dim: DimensionLike, /) -> str:
    return name_of(u.dimension(dim))


def can_broadcast_shapes(shapes: Sequence[Shape], /) -> bool:
    """Check if the shapes can be broadcasted together."""
    try:
        jnp.broadcast_shapes(*shapes)
    except ValueError:
        return False
    return True



"""Representation of coordinates in different systems."""

__all__ = ("AbstractVector", "is_vector")

from abc import abstractmethod
from collections.abc import Callable, Mapping
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, ClassVar, TypeGuard, TypeVar

import jax
import numpy as np
import quax_blocks
import wadler_lindig as wl

import quaxed.numpy as jnp
import unxt as u
from dataclassish import field_items, field_values, fields, replace

from .attribute import VectorAttribute
from .base import AbstractVectorLike
from .flags import AttrFilter
from coordinax._src.custom_types import Unit
from coordinax._src.utils import classproperty
from coordinax._src.vectors.api import vector
from coordinax._src.vectors.mixins import (
    AstropyRepresentationAPIMixin,
    IPythonReprMixin,
)

if TYPE_CHECKING:
    from typing import Self

VT = TypeVar("VT", bound="AbstractVector")


class AbstractVector(
    IPythonReprMixin,
    AstropyRepresentationAPIMixin,
    quax_blocks.NumpyInvertMixin[Any],
    quax_blocks.LaxLenMixin,
    AbstractVectorLike,
):
    """Base class for all vector types.

    A vector is a collection of components that can be represented in different
    coordinate systems. This class provides a common interface for all vector
    types. All fields of the vector are expected to be components of the vector.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    Scalar vectors have length 0:

    >>> vec = cx.vecs.CartesianPos1D.from_([1], "m")
    >>> len(vec)
    0

    Vectors with certain lengths:

    >>> vec = cx.vecs.CartesianPos1D(u.Quantity([1], "m"))
    >>> len(vec)
    1

    >>> vec = cx.vecs.CartesianPos1D(u.Quantity([1, 2], "m"))
    >>> len(vec)
    2

    String representation of vectors:

    - a vector with only axis fields

    >>> vec1 = cx.CartesianPos3D.from_([1, 2, 3], "m")
    >>> print(str(vec1))
    <CartesianPos3D: (x, y, z) [m]
        [1 2 3]>

    - a vector with additional attributes

    >>> vec2 = vec1.vconvert(cx.vecs.ProlateSpheroidalPos, Delta=u.Quantity(1, "m"))
    >>> print(str(vec2))
    <ProlateSpheroidalPos: (mu[m2], nu[m2], phi[rad])
        Delta=Quantity(1, unit='m')
        [14.374  0.626  1.107]>

    """

    # TODO: 1) a better variable name. 2) make this public (and frozen)
    #       3) a params_fields property that's better than `components`
    _AUX_FIELDS: ClassVar[tuple[str, ...]]

    def __init_subclass__(cls) -> None:
        """Determine properties of the vector class."""
        # Note that this is called before `dataclass` has had a chance to
        # process the class, so we cannot just call `fields(cls)` to get the
        # fields. Instead, we parse `cls.__annotation__` directly.

        # To find the auxiliary fields, we look for fields that are instances of
        # `VectorAttribute`. This is done by looking at the `__dict__` rather
        # than doing `getattr` since `VectorAttribute` is a descriptor and will
        # raise an error when accessed before the dataclass construction is
        # complete.
        aux = [
            k
            for k in cls.__annotations__
            if isinstance(cls.__dict__.get(k), VectorAttribute)
        ]
        cls._AUX_FIELDS = tuple(aux)

    # ===============================================================
    # Vector API

    # TODO: make public
    @abstractmethod
    def _dimensionality(self) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.vecs.CartesianPos2D._dimensionality()
        2

        """
        raise NotImplementedError  # pragma: no cover

    # TODO: a better name
    @property
    def _auxiliary_data(self) -> dict[str, Any]:
        """Get the auxiliary data of the vector.

        This is a dictionary of auxiliary data that is not part of the vector
        itself. This is used for storing additional information about the
        vector.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> vec = cx.vecs.CartesianPos3D.from_([1, 2, 3], "m")
        >>> vec._auxiliary_data
        {}

        >>> vec = cx.vecs.ProlateSpheroidalPos(
        ...     mu=u.Quantity(3, "m2"), nu=u.Quantity(2, "m2"),
        ...     phi=u.Quantity(4, "rad"), Delta=u.Quantity(1.5, "m"))
        >>> vec._auxiliary_data
        {'Delta': Quantity(Array(1.5, dtype=float32, ...), unit='m')}

        """
        return {k: getattr(self, k) for k in self._AUX_FIELDS}

    # ===============================================================
    # Array API

    # ---------------------------------------------------------------
    # Attributes

    # `.dtype`, `.shape`, `.size` handled by Quax
    # TODO: .device

    @property
    def mT(self) -> "Self":  # noqa: N802
        """Transpose the vector.

        Examples
        --------
        We assume the following imports:

        >>> import unxt as u
        >>> import coordinax as cx

        We can transpose a vector:

        >>> vec = cx.CartesianPos3D(x=u.Quantity([[0, 1], [2, 3]], "m"),
        ...                         y=u.Quantity([[0, 1], [2, 3]], "m"),
        ...                         z=u.Quantity([[0, 1], [2, 3]], "m"))
        >>> vec.mT.x
        Quantity(Array([[0, 2],
                                  [1, 3]], dtype=int32), unit='m')

        """
        return replace(self, **{k: v.mT for k, v in field_items(AttrFilter, self)})

    @property
    def ndim(self) -> int:
        """Number of array dimensions (axes).

        Examples
        --------
        We assume the following imports:

        >>> import unxt as u
        >>> import coordinax as cx

        We can get the number of dimensions of a vector:

        >>> vec = cx.vecs.CartesianPos2D.from_([1, 2], "m")
        >>> vec.ndim
        0

        >>> vec = cx.vecs.CartesianPos2D.from_([[1, 2], [3, 4]], "m")
        >>> vec.ndim
        1

        ``ndim`` is calculated from the broadcasted shape. We can
        see this by creating a 2D vector in which the components have
        different shapes:

        >>> vec = cx.vecs.CartesianPos2D(x=u.Quantity([[1, 2], [3, 4]], "m"),
        ...                              y=u.Quantity(0, "m"))
        >>> vec.ndim
        2

        """
        return len(self.shape)

    @property
    def T(self) -> "Self":  # noqa: N802
        """Transpose the vector.

        Examples
        --------
        We assume the following imports:

        >>> import unxt as u
        >>> import coordinax as cx

        We can transpose a vector:

        >>> vec = cx.CartesianPos3D(x=u.Quantity([[0, 1], [2, 3]], "m"),
        ...                         y=u.Quantity([[0, 1], [2, 3]], "m"),
        ...                         z=u.Quantity([[0, 1], [2, 3]], "m"))
        >>> vec.T.x
        Quantity(Array([[0, 2],
                                  [1, 3]], dtype=int32), unit='m')

        """
        return replace(self, **{k: v.T for k, v in field_items(AttrFilter, self)})

    # ===============================================================
    # Convenience methods

    def asdict(
        self, *, dict_factory: Callable[[Any], Mapping[str, u.AbstractQuantity]] = dict
    ) -> Mapping[str, u.AbstractQuantity]:
        """Return the vector as a Mapping.

        Parameters
        ----------
        dict_factory
            The type of the mapping to return.

        Returns
        -------
        Mapping[str, Quantity]
            The vector as a mapping.

        See Also
        --------
        `dataclasses.asdict`
            This applies recursively to the components of the vector.

        Examples
        --------
        We assume the following imports:

        >>> import unxt as u
        >>> import coordinax as cx

        We can get the vector as a mapping:

        >>> vec = cx.vecs.CartesianPos2D(x=u.Quantity([[1, 2], [3, 4]], "m"),
        ...                              y=u.Quantity(0, "m"))
        >>> vec.asdict()
        {'x': Quantity(Array([[1, 2], [3, 4]], dtype=int32), unit='m'),
         'y': Quantity(Array(0, dtype=int32, ...), unit='m')}

        """
        return dict_factory(field_items(self))

    @classproperty
    @classmethod
    def components(cls) -> tuple[str, ...]:
        """Vector component names.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.vecs.CartesianPos2D.components
        ('x', 'y')
        >>> cx.SphericalPos.components
        ('r', 'theta', 'phi')
        >>> cx.vecs.RadialVel.components
        ('r',)

        """
        return tuple(f.name for f in fields(AttrFilter, cls))

    @property
    def units(self) -> MappingProxyType[str, Unit]:
        """Get the units of the vector's components.

        Examples
        --------
        >>> import coordinax as cx
        >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "km")
        >>> vec.units
        mappingproxy({'x': Unit("km"), 'y': Unit("km"), 'z': Unit("km")})

        """
        return MappingProxyType(
            {k: u.unit_of(v) for k, v in field_items(AttrFilter, self)}
        )

    @classproperty
    @classmethod
    def dimensions(cls) -> dict[str, u.dims.AbstractDimension]:
        """Vector physical dimensions.

        Examples
        --------
        >>> import coordinax as cx

        >>> def pprint(d):
        ...     print({k: str(v).split("/")[0] for k, v in d.items()})

        >>> pprint(cx.vecs.CartesianPos1D.dimensions)
        {'x': 'length'}

        >>> pprint(cx.vecs.CartesianVel1D.dimensions)
        {'x': 'speed'}

        >>> pprint(cx.vecs.CartesianAcc1D.dimensions)
        {'x': 'acceleration'}

        >>> pprint(cx.vecs.RadialPos.dimensions)
        {'r': 'length'}

        >>> pprint(cx.vecs.RadialVel.dimensions)
        {'r': 'speed'}

        >>> pprint(cx.vecs.RadialAcc.dimensions)
        {'r': 'acceleration'}

        >>> pprint(cx.vecs.CartesianPos2D.dimensions)
        {'x': 'length', 'y': 'length'}

        >>> pprint(cx.vecs.CartesianVel2D.dimensions)
        {'x': 'speed', 'y': 'speed'}

        >>> pprint(cx.vecs.CartesianAcc2D.dimensions)
        {'x': 'acceleration', 'y': 'acceleration'}

        >>> pprint(cx.vecs.PolarPos.dimensions)
        {'r': 'length', 'phi': 'angle'}

        >>> pprint(cx.vecs.PolarVel.dimensions)
        {'r': 'speed', 'phi': 'angular frequency'}

        >>> pprint(cx.vecs.PolarAcc.dimensions)
        {'r': 'acceleration', 'phi': 'angular acceleration'}

        >>> pprint(cx.vecs.TwoSpherePos.dimensions)
        {'theta': 'angle', 'phi': 'angle'}

        >>> pprint(cx.vecs.TwoSphereVel.dimensions)
        {'theta': 'angular frequency', 'phi': 'angular frequency'}

        >>> pprint(cx.vecs.TwoSphereAcc.dimensions)
        {'theta': 'angular acceleration', 'phi': 'angular acceleration'}

        >>> pprint(cx.vecs.CartesianPos3D.dimensions)
        {'x': 'length', 'y': 'length', 'z': 'length'}

        >>> pprint(cx.vecs.CartesianVel3D.dimensions)
        {'x': 'speed', 'y': 'speed', 'z': 'speed'}

        >>> pprint(cx.vecs.CartesianAcc3D.dimensions)
        {'x': 'acceleration', 'y': 'acceleration', 'z': 'acceleration'}

        >>> pprint(cx.vecs.CylindricalPos.dimensions)
        {'rho': 'length', 'phi': 'angle', 'z': 'length'}

        >>> pprint(cx.vecs.CylindricalVel.dimensions)
        {'rho': 'speed', 'phi': 'angular frequency', 'z': 'speed'}

        >>> pprint(cx.vecs.CylindricalAcc.dimensions)
        {'rho': 'acceleration', 'phi': 'angular acceleration', 'z': 'acceleration'}

        >>> pprint(cx.vecs.SphericalPos.dimensions)
        {'r': 'length', 'theta': 'angle', 'phi': 'angle'}

        >>> pprint(cx.vecs.SphericalVel.dimensions)
        {'r': 'speed', 'theta': 'angular frequency', 'phi': 'angular frequency'}

        >>> pprint(cx.vecs.SphericalAcc.dimensions)
        {'r': 'acceleration', 'theta': 'angular acceleration', 'phi': 'angular acceleration'}

        >>> pprint(cx.vecs.LonLatSphericalPos.dimensions)
        {'lon': 'angle', 'lat': 'angle', 'distance': 'length'}

        >>> pprint(cx.vecs.LonLatSphericalVel.dimensions)
        {'lon': 'angular frequency', 'lat': 'angular frequency', 'distance': 'speed'}

        >>> pprint(cx.vecs.LonLatSphericalAcc.dimensions)
        {'lon': 'angular acceleration', 'lat': 'angular acceleration', 'distance': 'acceleration'}

        >>> pprint(cx.vecs.MathSphericalPos.dimensions)
        {'r': 'length', 'theta': 'angle', 'phi': 'angle'}

        >>> pprint(cx.vecs.MathSphericalVel.dimensions)
        {'r': 'speed', 'theta': 'angular frequency', 'phi': 'angular frequency'}

        >>> pprint(cx.vecs.MathSphericalAcc.dimensions)
        {'r': 'acceleration', 'theta': 'angular acceleration', 'phi': 'angular acceleration'}

        >>> pprint(cx.vecs.ProlateSpheroidalPos.dimensions)
        {'mu': 'area', 'nu': 'area', 'phi': 'angle'}

        >>> pprint(cx.vecs.ProlateSpheroidalVel.dimensions)
        {'mu': 'diffusivity', 'nu': 'diffusivity', 'phi': 'angular frequency'}

        >>> pprint(cx.vecs.ProlateSpheroidalAcc.dimensions)
        {'mu': 'dose of ionizing radiation', 'nu': 'dose of ionizing radiation', 'phi': 'angular acceleration'}

        >>> cx.vecs.Cartesian3D.dimensions
        <property object at ...>

        >>> cx.vecs.FourVector.dimensions
        <property object at ...>

        >>> pprint(cx.vecs.CartesianPosND.dimensions)
        {'q': 'length'}

        >>> pprint(cx.vecs.CartesianVelND.dimensions)
        {'q': 'speed'}

        >>> pprint(cx.vecs.CartesianAccND.dimensions)
        {'q': 'acceleration'}

        >>> pprint(cx.vecs.PoincarePolarVector.dimensions)
        {'rho': 'length', 'pp_phi': 'unknown', 'z': 'length',
         'dt_rho': 'speed', 'dt_pp_phi': 'unknown', 'dt_z': 'speed'}

        """  # noqa: E501
        return {f.name: u.dimension_of(f.type) for f in fields(AttrFilter, cls)}

    @property
    def dtypes(self) -> MappingProxyType[str, jnp.dtype[Any]]:
        """Get the dtypes of the vector's components.

        Examples
        --------
        >>> import coordinax as cx
        >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "km")
        >>> vec.dtypes
        mappingproxy({'x': dtype('int32'), 'y': dtype('int32'),
                      'z': dtype('int32')})

        """
        return MappingProxyType({k: v.dtype for k, v in field_items(AttrFilter, self)})

    @property
    def devices(self) -> MappingProxyType[str, jax.Device]:
        """Get the devices of the vector's components.

        Examples
        --------
        >>> import coordinax as cx
        >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "km")
        >>> vec.devices
        mappingproxy({'x': CpuDevice(id=0), 'y': CpuDevice(id=0),
                      'z': CpuDevice(id=0)})

        """
        return MappingProxyType({k: v.device for k, v in field_items(AttrFilter, self)})

    @property
    def shapes(self) -> MappingProxyType[str, tuple[int, ...]]:
        """Get the shapes of the vector's components.

        Examples
        --------
        >>> import coordinax as cx
        >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "km")
        >>> vec.shapes
        mappingproxy({'x': (), 'y': (), 'z': ()})

        """
        return MappingProxyType({k: v.shape for k, v in field_items(AttrFilter, self)})

    @property
    def sizes(self) -> MappingProxyType[str, int]:
        """Get the sizes of the vector's components.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.vecs.CartesianPos2D.from_([1, 2], "m").sizes
        mappingproxy({'x': 1, 'y': 1})

        >>> cx.vecs.CartesianPos2D.from_([[1, 2], [1, 2]], "m").sizes
        mappingproxy({'x': 2, 'y': 2})

        """
        return MappingProxyType({k: v.size for k, v in field_items(AttrFilter, self)})

    # ===============================================================
    # Wadler-Lindig

    def _pdoc_comps(self) -> wl.AbstractDoc:
        # make the components string
        units_ = self.units
        if len(set(units_.values())) == 1:
            cdocs = [wl.TextDoc(f"{c}") for c in self.components]
            end = wl.TextDoc(f") [{units_[self.components[0]]}]")
        else:
            cdocs = [wl.TextDoc(f"{c}[{units_[c]}]") for c in self.components]
            end = wl.TextDoc(")")
        return wl.bracketed(
            begin=wl.TextDoc("("), docs=cdocs, sep=wl.comma, end=end, indent=4
        )

    def __pdoc__(self, *, vector_form: bool = False, **kwargs: Any) -> wl.AbstractDoc:
        """Return the Wadler-Lindig docstring for the vector.

        Parameters
        ----------
        vector_form
            If True, return the vector form of the docstring.
        short_arrays
            If True, use short arrays for the docstring.
        **kwargs
            Additional keyword arguments to pass to the Wadler-Lindig docstring
            formatter.

        """
        if not vector_form:
            # TODO: not use private API.
            return wl._definitions._pformat_dataclass(self, **kwargs)

        # -----------------------------

        cls_name = wl.TextDoc(self.__class__.__name__)

        # make the components string
        comps_doc = self._pdoc_comps()

        # make the aux fields string
        if not self._AUX_FIELDS:
            aux_doc = wl.TextDoc("")
        else:
            aux_doc = (
                wl.TextDoc("\n")  # force a line break
                + wl.bracketed(
                    begin=wl.TextDoc(" "),  # indent to opening "<"
                    docs=wl.named_objs(
                        [(k, getattr(self, k)) for k in self._AUX_FIELDS],
                        **kwargs | {"short_arrays": "compact"},
                    ),
                    sep=wl.comma,
                    end=wl.TextDoc(""),  # no end bracket
                    indent=1,
                )
            )

        # make the values string
        # TODO: avoid casting to numpy arrays
        fvals = tuple(
            map(u.ustrip, jnp.broadcast_arrays(*field_values(AttrFilter, self)))
        )
        fvs = np.array2string(
            np.stack(fvals, axis=-1),
            precision=kwargs.pop("precision", 3),
            threshold=kwargs.pop("threshold", 1000),
            suffix=">",
        )
        vs_doc = wl.TextDoc(fvs).nest(4)

        # TODO: figure out how to avoid the hacky line breaks
        return (
            (  # header
                (wl.TextDoc("<") + cls_name + wl.TextDoc(":")).group()
                + wl.BreakDoc(" ")
                + comps_doc
            ).group()
            + aux_doc.group()  # aux fields
            + wl.BreakDoc(" " * 9_999)  # TODO: better way to force break
            + (  # values (including end >)
                wl.TextDoc("    ") + vs_doc + wl.TextDoc(">")
            ).group()
        )


# ---------------------------------------------------------------
# Constructors


@AbstractVector.from_.dispatch  # type: ignore[untyped-decorator]
def from_(cls: type[AbstractVector], *args: Any, **kwargs: Any) -> AbstractVector:
    """Create a vector from arguments.

    See `coordinax.vector` for more information.

    """
    return vector(cls, *args, **kwargs)


# ================================================================


def is_vector(obj: Any, /) -> TypeGuard[AbstractVector]:
    """Check if the object is a `AbstractVector` object.

    Examples
    --------
    >>> import coordinax as cx

    >>> vec = cx.vecs.CartesianPos3D.from_([1, 2, 3], "m")
    >>> cx.vecs.is_vector(vec)
    True

    >>> space = cx.vecs.KinematicSpace.from_(vec)
    >>> cx.vecs.is_vector(space)
    False

    >>> coord = cx.Coordinate.from_(vec, cx.frames.ICRS())
    >>> cx.vecs.is_vector(coord)
    True

    >>> cx.vecs.is_vector(42)
    False

    """
    return isinstance(obj, AbstractVector)  # pragma: no cover



"""N-dimensional."""

__all__ = (
    # Base
    "AbstractPosND",
    "AbstractVelND",
    "AbstractAccND",
    # Cartesian
    "CartesianPosND",
    "CartesianVelND",
    "CartesianAccND",
    # Poincare
    "PoincarePolarVector",
)

from .base import AbstractAccND, AbstractPosND, AbstractVelND
from .cartesian import CartesianAccND, CartesianPosND, CartesianVelND
from .poincare import PoincarePolarVector

# isort: split
from . import (
    register_plum,  # noqa: F401
    register_primitives,  # noqa: F401
    register_unxt,  # noqa: F401
    register_vectorapi,  # noqa: F401
)



"""Built-in 4-vector classes."""

__all__: tuple[str, ...] = ()


import jax.numpy as jnp
from jaxtyping import Shaped
from plum import conversion_method, convert

import quaxed.numpy as jnp
import unxt as u
from unxt.quantity import BareQuantity

from .spacetime import FourVector
from coordinax._src.vectors.d3 import (
    CartesianPos3D,
    CylindricalPos,
    LonLatSphericalPos,
    MathSphericalPos,
    SphericalPos,
)


@conversion_method(type_from=FourVector, type_to=BareQuantity)
def fourvec_to_quantity(obj: FourVector, /) -> Shaped[BareQuantity, "*batch 4"]:
    """`coordinax.AbstractPos3D` -> `unxt.quantity.BareQuantity`.

    Convert the 4-vector to a Quantity array with the components as the last
    dimension.

    Examples
    --------
    >>> from plum import convert
    >>> import unxt as u
    >>> import coordinax as cx

    >>> w = cx.vecs.FourVector (t=u.Quantity([1, 2], "yr"),
    ...                        q=u.Quantity([[1, 2, 3], [4, 5, 6]], "pc"))

    >>> convert(w, u.quantity.BareQuantity).uconvert("pc")
    BareQuantity(Array([[0.3066014, 1. , 2. , 3. ],
                        [0.6132028, 4. , 4.9999995, 6. ]],
                       dtype=float32, ...),
                 unit='pc')

    """
    cart = convert(obj.q, BareQuantity)
    ct = convert(obj.c * obj.t[..., None], BareQuantity)
    return jnp.concat([ct, cart], axis=-1)


@conversion_method(type_from=FourVector, type_to=u.Quantity)
def fourvec_to_quantity(obj: FourVector, /) -> Shaped[u.Quantity["length"], "*batch 4"]:
    """`coordinax.AbstractPos3D` -> `unxt.Quantity`.

    Convert the 4-vector to a Quantity array with the components as the last
    dimension.

    Examples
    --------
    >>> from plum import convert
    >>> import unxt as u
    >>> import coordinax as cx

    >>> w = cx.vecs.FourVector (t=u.Quantity([1, 2], "yr"),
    ...                        q=u.Quantity([[1, 2, 3], [4, 5, 6]], "pc"))

    >>> convert(w, u.Quantity).uconvert("pc")
    Quantity(Array([[0.3066014, 1. , 2. , 3. ],
                              [0.6132028, 4. , 4.9999995, 6. ]],
                             dtype=float32, ...),
                       unit='pc')

    """
    cart: u.Quantity = convert(obj.q, u.Quantity)
    ct = obj.c * obj.t[..., None]
    return jnp.concat([ct, cart], axis=-1)


@conversion_method(type_from=FourVector, type_to=CartesianPos3D)  # type: ignore[arg-type]
def convert_4vec_to_cart3d(obj: FourVector, /) -> CartesianPos3D:
    """Convert a 4-vector to a Cartesian 3-vector.

    Examples
    --------
    >>> from plum import convert
    >>> import unxt as u
    >>> import coordinax as cx

    >>> w = cx.FourVector (t=u.Quantity(1, "s"), q=u.Quantity([1, 2, 3], "m"))
    >>> print(convert(w, cx.vecs.CartesianPos3D))
    <CartesianPos3D: (x, y, z) [m]
        [1 2 3]>

    """
    return convert(obj.q, CartesianPos3D)


@conversion_method(type_from=FourVector, type_to=CylindricalPos)  # type: ignore[arg-type]
def convert_4vec_to_cylindrical(obj: FourVector, /) -> CylindricalPos:
    """Convert a 4-vector to a Cylindrical 3-vector.

    Examples
    --------
    >>> from plum import convert
    >>> import unxt as u
    >>> import coordinax as cx

    >>> w = cx.FourVector (t=u.Quantity(1, "s"), q=u.Quantity([1, 2, 3], "m"))
    >>> print(convert(w, cx.vecs.CylindricalPos))
    <CylindricalPos: (rho[m], phi[rad], z[m])
        [2.236 1.107 3.   ]>

    """
    return convert(obj.q, CylindricalPos)


@conversion_method(type_from=FourVector, type_to=SphericalPos)  # type: ignore[arg-type]
def convert_4vec_to_spherical(obj: FourVector, /) -> SphericalPos:
    """Convert a 4-vector to a spherical 3-vector.

    Examples
    --------
    >>> from plum import convert
    >>> import unxt as u
    >>> import coordinax as cx

    >>> w = cx.FourVector (t=u.Quantity(1, "s"), q=u.Quantity([1, 2, 3], "m"))
    >>> print(convert(w, cx.SphericalPos))
    <SphericalPos: (r[m], theta[rad], phi[rad])
        [3.742 0.641 1.107]>

    """
    return convert(obj.q, SphericalPos)


@conversion_method(type_from=FourVector, type_to=LonLatSphericalPos)  # type: ignore[arg-type]
def convert_4vec_to_lonlat_spherical(obj: FourVector, /) -> LonLatSphericalPos:
    """Convert a 4-vector to a lon-lat spherical 3-vector.

    Examples
    --------
    >>> from plum import convert
    >>> import unxt as u
    >>> import coordinax as cx

    >>> w = cx.FourVector (t=u.Quantity(1, "s"), q=u.Quantity([1, 2, 3], "m"))
    >>> print(convert(w, cx.vecs.LonLatSphericalPos))
    <LonLatSphericalPos: (lon[rad], lat[deg], distance[m])
        [ 1.107 53.301  3.742]>

    """
    return convert(obj.q, LonLatSphericalPos)


@conversion_method(type_from=FourVector, type_to=MathSphericalPos)  # type: ignore[arg-type]
def convert_4vec_to_mathsph(obj: FourVector, /) -> MathSphericalPos:
    """Convert a 4-vector to a math spherical 3-vector.

    Examples
    --------
    >>> from plum import convert
    >>> import unxt as u
    >>> import coordinax as cx

    >>> w = cx.FourVector (t=u.Quantity(1, "s"), q=u.Quantity([1, 2, 3], "m"))
    >>> print(convert(w, cx.vecs.MathSphericalPos))
    <MathSphericalPos: (r[m], theta[rad], phi[rad])
        [3.742 1.107 0.641]>

    """
    return convert(obj.q, MathSphericalPos)



"""Base implementation of coordinate frames."""

__all__ = ("Alice", "FriendOfAlice", "Bob")


from typing import final

from plum import dispatch

import unxt as u

from .base import AbstractReferenceFrame
from coordinax._src.frames import api
from coordinax._src.operators import (
    GalileanBoost,
    GalileanRotation,
    GalileanSpatialTranslation,
    Identity,
    Pipe,
)


@final
class Alice(AbstractReferenceFrame):
    """A stationary lab reference frame.

    Examples
    --------
    >>> import coordinax.frames as cxf

    >>> alice = cxf.Alice()
    >>> bob = cxf.Bob()

    >>> op = cxf.frame_transform_op(alice, bob)
    >>> print(op)
    Pipe((
      GalileanSpatialTranslation(<CartesianPos3D: (x, y, z) [km]
          [100000  10000      0]>),
      GalileanBoost(<CartesianVel3D: (x, y, z) [m / s]
          [2.698e+08 0.000e+00 0.000e+00]>)
    ))

    """


@final
class FriendOfAlice(AbstractReferenceFrame):
    """A reference frame shifted from Alice's frame."""


@final
class Bob(AbstractReferenceFrame):
    """A moving reference frame.

    Examples
    --------
    >>> import coordinax.frames as cxf

    >>> alice = cxf.Alice()
    >>> bob = cxf.Bob()

    >>> op = cxf.frame_transform_op(alice, bob)
    >>> print(op)
    Pipe((
      GalileanSpatialTranslation(<CartesianPos3D: (x, y, z) [km]
          [100000  10000      0]>),
      GalileanBoost(<CartesianVel3D: (x, y, z) [m / s]
          [2.698e+08 0.000e+00 0.000e+00]>)
    ))

    """


# ===================================================================


@dispatch.multi((Alice, Alice), (FriendOfAlice, FriendOfAlice), (Bob, Bob))
def frame_transform_op(
    from_frame: AbstractReferenceFrame,
    to_frame: AbstractReferenceFrame,
    /,
) -> Identity:
    """Return an identity operator for frames that are the same.

    Examples
    --------
    >>> import coordinax.frames as cxf

    >>> alice = cxf.Alice()
    >>> cxf.frame_transform_op(alice, alice)
    Identity()

    >>> friend = cxf.FriendOfAlice()
    >>> cxf.frame_transform_op(friend, friend)
    Identity()

    >>> bob = cxf.Bob()
    >>> cxf.frame_transform_op(bob, bob)
    Identity()

    """
    return Identity()


@dispatch
def frame_transform_op(from_frame: Alice, to_frame: FriendOfAlice, /) -> Pipe:
    """Transform from Alice's frame to Bob's frame.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> alice = cx.frames.Alice()
    >>> friend = cx.frames.FriendOfAlice()

    >>> op = cx.frames.frame_transform_op(alice, friend)
    >>> print(op)
    Pipe((
      GalileanSpatialTranslation(<CartesianPos3D: (x, y, z) [m]
          [10  0  0]>),
      GalileanRotation([[ 0.         -0.99999994  0.        ]
                        [ 0.99999994  0.          0.        ]
                        [ 0.          0.          0.99999994]])
    ))

    >>> q_alice = cx.vecs.CartesianPos3D.from_([0, 0, 0], "m")
    >>> q_friend = op(u.Quantity(1, "s"), q_alice)
    >>> q_friend
    (Quantity(Array(1, dtype=int32, weak_type=True), unit='s'),
     CartesianPos3D( x=Quantity(0., unit='m'), y=Quantity(9.999999, unit='m'),
                     z=Quantity(0., unit='m')
     ))

    """
    shift = GalileanSpatialTranslation.from_([10, 0, 0], "m")
    rotation = GalileanRotation.from_euler("Z", u.Quantity(90, "deg"))
    return shift | rotation


@dispatch
def frame_transform_op(from_frame: Alice, to_frame: Bob, /) -> Pipe:
    """Transform from Alice's frame to Bob's frame.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv
    >>> import coordinax.frames as cxf

    >>> alice = cxf.Alice()
    >>> bob = cxf.Bob()

    >>> op = cxf.frame_transform_op(alice, bob)
    >>> print(op)
    Pipe((
      GalileanSpatialTranslation(<CartesianPos3D: (x, y, z) [km]
          [100000  10000      0]>),
      GalileanBoost(<CartesianVel3D: (x, y, z) [m / s]
          [2.698e+08 0.000e+00 0.000e+00]>)
    ))

    >>> q = cxv.CartesianPos3D.from_([0, 0, 0], "m")

    >>> q_bob = op(u.Quantity(0, "s"), q)
    >>> q_bob
    (Quantity(Array(0, dtype=int32, weak_type=True), unit='s'),
     CartesianPos3D( x=Quantity(1.e+08, unit='m'), ... z=Quantity(0., unit='m')
    ))

    >>> q_bob = op(u.Quantity(1, "s"), q)
    >>> q_bob
    (Quantity(Array(1, dtype=int32, weak_type=True), unit='s'),
     CartesianPos3D( x=Quantity(3.6981322e+08, unit='m'), ... z=Quantity(0., unit='m')
    ))

    """
    shift = GalileanSpatialTranslation.from_([100_000, 10_000, 0], "km")
    boost = GalileanBoost.from_([269_813_212.2, 0, 0], "m/s")
    return shift | boost


@dispatch.multi(
    (FriendOfAlice, Alice),
    (Bob, Alice),
)
def frame_transform_op(
    from_frame: AbstractReferenceFrame, to_frame: AbstractReferenceFrame, /
) -> Pipe:
    """Transform back.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> alice = cx.frames.Alice()
    >>> friend = cx.frames.FriendOfAlice()

    >>> op = cx.frames.frame_transform_op(friend, alice)
    >>> print(op)
    Pipe((
      GalileanRotation([[ 0.          0.99999994  0.        ]
                        [-0.99999994  0.          0.        ]
                        [ 0.          0.          0.99999994]]),
      GalileanSpatialTranslation(<CartesianPos3D: (x, y, z) [m]
          [-10   0   0]>)
    ))

    >>> q_friend = cx.vecs.CartesianPos3D.from_([0, 0, 0], "m")
    >>> q_alice = op(u.Quantity(1, "s"), q_friend)
    >>> q_alice
    (Quantity(Array(1, dtype=int32, weak_type=True), unit='s'),
     CartesianPos3D( x=Quantity(-10., unit='m'), y=Quantity(0., unit='m'),
                     z=Quantity(0., unit='m')
    ))

    """
    return api.frame_transform_op(to_frame, from_frame).inverse  # pylint: disable=W1114



"""Radial 1D vectors."""

__all__ = ("RadialAcc", "RadialPos", "RadialVel")

from typing import final

import equinox as eqx

import unxt as u
from dataclassish.converters import Unless

import coordinax._src.custom_types as ct
from .base import AbstractAcc1D, AbstractPos1D, AbstractVel1D
from coordinax._src.distances import AbstractDistance, BatchableDistance, Distance


@final
class RadialPos(AbstractPos1D):
    """Radial vector representation.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.vecs.RadialPos(u.Quantity([2], "m"))
    >>> print(vec)
    <RadialPos: (r) [m]
        [[2]]>

    """

    r: BatchableDistance = eqx.field(converter=Unless(AbstractDistance, Distance.from_))
    r"""Radial distance :math:`r \in [0,+\infty)`."""


@final
class RadialVel(AbstractVel1D):
    """Radial velocity.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.vecs.RadialVel(u.Quantity([2], "m/s"))
    >>> print(vec)
    <RadialVel: (r) [m / s]
        [[2]]>

    """

    r: ct.BBtSpeed = eqx.field(converter=u.Quantity["speed"].from_)
    r"""Radial speed :math:`dr/dt \in (-\infty,+\infty)`."""


@final
class RadialAcc(AbstractAcc1D):
    """Radial Acceleration.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.vecs.RadialAcc(u.Quantity([2], "m/s2"))
    >>> print(vec)
    <RadialAcc: (r) [m / s2]
        [[2]]>

    """

    r: ct.BBtAcc = eqx.field(converter=u.Quantity["acceleration"].from_)
    r"""Radial acceleration :math:`d^2r/dt^2 \in (-\infty,+\infty)`."""



"""Built-in 4-vector classes."""

__all__: tuple[str, ...] = ()

from dataclasses import replace
from typing import Any, cast

import equinox as eqx
from plum import dispatch

import unxt as u

from .spacetime import FourVector
from coordinax._src.vectors import api
from coordinax._src.vectors.d3 import AbstractPos3D, CartesianPos3D


@dispatch
def vector(cls: type[FourVector], obj: u.AbstractQuantity, /) -> FourVector:
    """Construct a vector from a Quantity array.

    The ``Quantity[Any, (*#batch, 4), "..."]`` is expected to have the
    components as the last dimension. The 4 components are the (c x) time, x, y,
    z.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    >>> xs = u.Quantity([0, 1, 2, 3], "meter")  # [ct, x, y, z]
    >>> vec = cx.FourVector.from_(xs)
    >>> print(vec)
    <FourVector: (t[m s / km], q=(x, y, z) [m])
        [0. 1. 2. 3.]>

    >>> xs = u.Quantity(jnp.array([[0, 1, 2, 3], [10, 4, 5, 6]]), "meter")
    >>> vec = cx.FourVector.from_(xs)
    >>> print(vec)
    <FourVector: (t[m s / km], q=(x, y, z) [m])
        [[0.000e+00 1.000e+00 2.000e+00 3.000e+00]
         [3.336e-05 4.000e+00 5.000e+00 6.000e+00]]>

    """
    _ = eqx.error_if(
        obj,
        obj.shape[-1] != 4,
        f"Cannot construct {cls} from array with shape {obj.shape}.",
    )
    c = cls.__dataclass_fields__["c"].default.default
    return cls(t=obj[..., 0] / c, q=obj[..., 1:], c=c)


# =============================================================================


@dispatch
def vconvert(
    spatial_target: type[AbstractPos3D], current: FourVector, /, **kwargs: Any
) -> FourVector:
    """Convert the spatial part of a 4-vector to a different 3-vector.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> w = cx.FourVector (t=u.Quantity(1, "s"), q=u.Quantity([1, 2, 3], "m"))
    >>> print(cx.vconvert(cx.vecs.CylindricalPos, w))
    <FourVector: (t[s], q=(rho[m], phi[rad], z[m]))
        [1.    2.236 1.107 3.   ]>

    """
    q = cast("AbstractPos3D", api.vconvert(spatial_target, current.q, **kwargs))
    return replace(current, q=q)


# =============================================================================


@dispatch
def spatial_component(x: FourVector, /) -> AbstractPos3D:
    """Return the spatial component of the vector.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> w = cx.FourVector (t=u.Quantity(1, "s"), q=u.Quantity([1, 2, 3], "m"))
    >>> print(spatial_component(w))
    <CartesianPos3D: (x, y, z) [m]
        [1 2 3]>

    """
    return x.q


# =============================================================================
# Corresponding Cartesian class


@dispatch
def cartesian_vector_type(
    obj: type[FourVector] | FourVector, /
) -> type[CartesianPos3D]:
    """Return the corresponding Cartesian vector class."""
    return CartesianPos3D



"""Register primitives for Space."""

__all__: tuple[str, ...] = ()


from typing import Any

import jax
from quax import register

import quaxed.numpy as jnp
from dataclassish import replace

from .core import KinematicSpace
from coordinax._src.custom_types import Shape


@register(jax.lax.broadcast_in_dim_p)
def broadcast_in_dim_p_space(  # TODO: KinematicSpace[PosT] -- plum#212
    obj: KinematicSpace, /, *, shape: Shape, **kw: Any
) -> KinematicSpace:
    """Broadcast in a dimension."""
    batch = shape[:-1]
    return replace(
        obj,
        **{
            k: jnp.broadcast_to(v, (*batch, v.aval().shape[-1])) for k, v in obj.items()
        },
    )


@register(jax.lax.neg_p)  # TODO: KinematicSpace[PosT] -- plum#212
def neg_p_space(space: KinematicSpace, /) -> KinematicSpace:
    """Negative of the vector.

    Examples
    --------
    >>> import coordinax as cx

    >>> w = cx.KinematicSpace(
    ...     length=cx.CartesianPos3D.from_([[[1, 2, 3], [4, 5, 6]]], "m"),
    ...     speed=cx.CartesianVel3D.from_([[[1, 2, 3], [4, 5, 6]]], "m/s")
    ... )

    >>> print(-w)
    KinematicSpace({
       'length': <CartesianPos3D: (x, y, z) [m]
           [[[-1 -2 -3]
             [-4 -5 -6]]]>,
       'speed': <CartesianVel3D: (x, y, z) [m / s]
           [[[-1 -2 -3]
             [-4 -5 -6]]]>
    })

    """
    return type(space)(**{k: -v for k, v in space.items()})



"""Type hints for angles."""

__all__ = ("BatchableAngle", "BatchableAngleQ")


from jaxtyping import Shaped

import unxt as u

#: Batchable angular-type Quantity.
BatchableAngularQuantity = Shaped[u.Quantity["angle"], "*#batch"]

#: Batchable Angle.
BatchableAngle = Shaped[u.quantity.AbstractAngle, "*#batch"]

#: Batchable Angle or Angular Quantity.
BatchableAngleQ = BatchableAngle | BatchableAngularQuantity



"""Representation of coordinates in different systems."""

__all__: tuple[str, ...] = ()

import functools as ft

import jax
from plum import dispatch

import quaxed.numpy as jnp

import coordinax._src.vectors.custom_types as ct
from .base import AbstractAcc2D, AbstractPos2D, AbstractVel2D
from .cartesian import CartesianAcc2D, CartesianPos2D, CartesianVel2D
from .polar import PolarAcc, PolarPos, PolarVel
from .spherical import TwoSphereAcc, TwoSpherePos, TwoSphereVel
from coordinax._src.vectors.private_api import combine_aux, wrap_vconvert_impl_params

###############################################################################
# Vector Transformation


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
def vconvert(
    to_vector: type[AbstractPos2D],
    from_vector: type[AbstractPos2D],
    params: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """AbstractPos -> CartesianPos1D -> AbstractPos."""
    params, aux = vconvert(
        CartesianPos2D, from_vector, params, in_aux=in_aux, out_aux=None, units=units
    )
    params, aux = vconvert(
        to_vector, CartesianPos2D, params, in_aux=aux, out_aux=out_aux, units=units
    )
    return params, aux


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[PolarPos],
    from_vector: type[CartesianPos2D],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.OptAuxDict]:
    """CartesianPos2D -> PolarPos.

    The `x` and `y` coordinates are converted to the radial coordinate `r` and
    the angular coordinate `phi`.

    """
    r = jnp.hypot(p["x"], p["y"])
    phi = jnp.atan2(p["y"], p["x"])
    return {"r": r, "phi": phi}, combine_aux(in_aux, out_aux)


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[CartesianPos2D],
    from_vector: type[PolarPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.OptAuxDict]:
    """PolarPos -> CartesianPos2D.

    The `r` and `phi` coordinates are converted to the `x` and `y` coordinates.

    """
    x = p["r"] * jnp.cos(p["phi"])
    y = p["r"] * jnp.sin(p["phi"])
    return {"x": x, "y": y}, combine_aux(in_aux, out_aux)


###############################################################################
# Corresponding Cartesian classes


@dispatch
def cartesian_vector_type(
    obj: type[AbstractPos2D] | AbstractPos2D, /
) -> type[CartesianPos2D]:
    """AbstractPos2D -> CartesianPos2D."""
    return CartesianPos2D


@dispatch
def cartesian_vector_type(
    obj: type[AbstractVel2D] | AbstractVel2D, /
) -> type[CartesianVel2D]:
    """AbstractVel2D -> CartesianVel2D."""
    return CartesianVel2D


@dispatch
def cartesian_vector_type(
    obj: type[AbstractAcc2D] | AbstractAcc2D, /
) -> type[CartesianAcc2D]:
    """AbstractPos -> CartesianAcc2D."""
    return CartesianAcc2D


###############################################################################
# Corresponding time derivative classes

# -----------------------------------------------
# Position -> Velocity


@dispatch
def time_derivative_vector_type(
    obj: type[CartesianPos2D] | CartesianPos2D, /
) -> type[CartesianVel2D]:
    """Return the corresponding time derivative class."""
    return CartesianVel2D


@dispatch
def time_derivative_vector_type(obj: type[PolarPos] | PolarPos, /) -> type[PolarVel]:
    """Return the corresponding time derivative class."""
    return PolarVel


@dispatch
def time_derivative_vector_type(
    obj: type[TwoSpherePos] | TwoSpherePos, /
) -> type[TwoSphereVel]:
    """Return the corresponding time derivative class."""
    return TwoSphereVel


# -----------------------------------------------
# Velocity -> Position


@dispatch
def time_antiderivative_vector_type(
    obj: type[CartesianVel2D] | CartesianVel2D, /
) -> type[CartesianPos2D]:
    """Return the corresponding time antiderivative class."""
    return CartesianPos2D


@dispatch
def time_antiderivative_vector_type(
    obj: type[PolarVel] | PolarVel, /
) -> type[PolarPos]:
    """Return the corresponding time antiderivative class."""
    return PolarPos


@dispatch
def time_antiderivative_vector_type(
    obj: type[TwoSphereVel] | TwoSphereVel, /
) -> type[TwoSpherePos]:
    """Return the corresponding time antiderivative class."""
    return TwoSpherePos


# -----------------------------------------------
# Velocity -> Acceleration


@dispatch
def time_derivative_vector_type(
    obj: type[CartesianVel2D] | CartesianVel2D, /
) -> type[CartesianAcc2D]:
    """Return the corresponding time derivative class."""
    return CartesianAcc2D


@dispatch
def time_derivative_vector_type(obj: type[PolarVel] | PolarVel, /) -> type[PolarAcc]:
    """Return the corresponding time derivative class."""
    return PolarAcc


@dispatch
def time_derivative_vector_type(
    obj: type[TwoSphereVel] | TwoSphereVel, /
) -> type[TwoSphereAcc]:
    """Return the corresponding time derivative class."""
    return TwoSphereAcc


# -----------------------------------------------
# Acceleration -> Velocity


@dispatch
def time_antiderivative_vector_type(
    obj: type[CartesianAcc2D] | CartesianAcc2D, /
) -> type[CartesianVel2D]:
    """Return the corresponding time antiderivative class."""
    return CartesianVel2D


@dispatch
def time_antiderivative_vector_type(
    obj: type[PolarAcc] | PolarAcc, /
) -> type[PolarVel]:
    """Return the corresponding time antiderivative class."""
    return PolarVel


@dispatch
def time_antiderivative_vector_type(
    obj: type[TwoSphereAcc] | TwoSphereAcc, /
) -> type[TwoSphereVel]:
    """Return the corresponding time antiderivative class."""
    return TwoSphereVel



"""Built-in vector classes."""

__all__ = (
    "CartesianAcc3D",
    "CartesianPos3D",
    "CartesianVel3D",
)

import functools as ft
from typing import final
from typing_extensions import override

import equinox as eqx
import jax

import quaxed.numpy as jnp
import unxt as u

import coordinax._src.custom_types as ct
from .base import AbstractAcc3D, AbstractPos3D, AbstractVel3D
from coordinax._src.distances import BBtLength
from coordinax._src.vectors.base.cartesian import AbstractCartesian


@final
class CartesianPos3D(AbstractCartesian, AbstractPos3D):
    """Cartesian 3D Position.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.CartesianPos3D.from_(u.Quantity([1, 2, 3], "m"))
    >>> print(vec)
    <CartesianPos3D: (x, y, z) [m]
        [1 2 3]>

    """

    x: BBtLength = eqx.field(converter=u.Quantity["length"].from_)
    r"""X coordinate :math:`x \in (-\infty,+\infty)`."""

    y: BBtLength = eqx.field(converter=u.Quantity["length"].from_)
    r"""Y coordinate :math:`y \in (-\infty,+\infty)`."""

    z: BBtLength = eqx.field(converter=u.Quantity["length"].from_)
    r"""Z coordinate :math:`z \in (-\infty,+\infty)`."""


@final
class CartesianVel3D(AbstractCartesian, AbstractVel3D):
    """Cartesian 3D Velocity.

    Examples
    --------
    >>> import coordinax as cx

    >>> vec = cx.CartesianVel3D.from_([1, 2, 3], "m/s")
    >>> print(vec)
    <CartesianVel3D: (x, y, z) [m / s]
        [1 2 3]>

    """

    x: ct.BBtSpeed = eqx.field(converter=u.Quantity["speed"].from_)
    r"""X speed :math:`dx/dt \in [-\infty, \infty]."""

    y: ct.BBtSpeed = eqx.field(converter=u.Quantity["speed"].from_)
    r"""Y speed :math:`dy/dt \in [-\infty, \infty]."""

    z: ct.BBtSpeed = eqx.field(converter=u.Quantity["speed"].from_)
    r"""Z speed :math:`dz/dt \in [-\infty, \infty]."""

    @ft.partial(eqx.filter_jit, inline=True)
    def norm(self, _: AbstractPos3D | None = None, /) -> ct.BBtSpeed:  # type: ignore[misc]
        """Return the norm of the vector.

        Examples
        --------
        >>> import coordinax as cx
        >>> c = cx.CartesianVel3D.from_([1, 2, 3], "km/s")
        >>> c.norm()
        Quantity(Array(3.7416575, dtype=float32), unit='km / s')

        """
        return jnp.sqrt(self.x**2 + self.y**2 + self.z**2)


@final
class CartesianAcc3D(AbstractCartesian, AbstractAcc3D):
    """Cartesian differential representation."""

    x: ct.BBtAcc = eqx.field(converter=u.Quantity["acceleration"].from_)
    r"""X acceleration :math:`d^2x/dt^2 \in [-\infty, \infty]."""

    y: ct.BBtAcc = eqx.field(converter=u.Quantity["acceleration"].from_)
    r"""Y acceleration :math:`d^2y/dt^2 \in [-\infty, \infty]."""

    z: ct.BBtAcc = eqx.field(converter=u.Quantity["acceleration"].from_)
    r"""Z acceleration :math:`d^2z/dt^2 \in [-\infty, \infty]."""

    @override
    @ft.partial(jax.jit, inline=True)
    def norm(
        self, _: AbstractVel3D | None = None, __: AbstractPos3D | None = None, /
    ) -> ct.BBtAcc:
        """Return the norm of the vector.

        Examples
        --------
        >>> import coordinax as cx
        >>> c = cx.vecs.CartesianAcc3D.from_([1, 2, 3], "km/s2")
        >>> c.norm()
        Quantity(Array(3.7416575, dtype=float32), unit='km / s2')

        """
        return jnp.sqrt(self.x**2 + self.y**2 + self.z**2)



"""Transformations for Vectors."""

__all__: tuple[str, ...] = ()

from .collection import *
from .common import *
from .d1 import *
from .d2 import *
from .d3 import *
from .dn import *
from .time_derivatives import *



"""Built-in vector classes."""

__all__: tuple[str, ...] = ()


from jaxtyping import Shaped
from plum import conversion_method

import unxt as u

from .cartesian import CartesianPosND


@conversion_method(CartesianPosND, u.Quantity)  # type: ignore[arg-type]
def vec_to_q(obj: CartesianPosND, /) -> Shaped[u.Quantity["length"], "*batch N"]:
    """`coordinax.AbstractPos3D` -> `unxt.Quantity`.

    Examples
    --------
    >>> from plum import convert
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.vecs.CartesianPosND(u.Quantity([1, 2, 3, 4, 5], unit="km"))
    >>> convert(vec, u.Quantity)
    Quantity(Array([1, 2, 3, 4, 5], dtype=int32), unit='km')

    """
    return obj.q



"""Representation of coordinates in different systems."""

__all__ = ("AbstractAcc3D", "AbstractPos3D", "AbstractVel3D")

from coordinax._src.vectors.base_acc import AbstractAcc
from coordinax._src.vectors.base_pos import AbstractPos
from coordinax._src.vectors.base_vel import AbstractVel


class AbstractPos3D(AbstractPos):
    """Abstract representation of 3D coordinates in different systems."""

    @classmethod
    def _dimensionality(cls) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.vecs.CartesianPos3D._dimensionality()
        3

        """
        return 3


class AbstractVel3D(AbstractVel):
    """Abstract representation of 3D vector differentials."""

    @classmethod
    def _dimensionality(cls) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.vecs.CartesianVel3D._dimensionality()
        3

        """
        return 3


class AbstractAcc3D(AbstractAcc):
    """Abstract representation of 3D vector accelerations."""

    @classmethod
    def _dimensionality(cls) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.vecs.CartesianAcc3D._dimensionality()
        3

        """
        return 3



"""Register dispatches for `unxt`."""

__all__: tuple[str, ...] = ()


from plum import dispatch

from dataclassish import replace

from .base import AbstractReferenceFrame
from .coordinate import Coordinate
from .xfm import TransformedReferenceFrame
from coordinax._src.operators import AbstractOperator
from coordinax._src.vectors.base_pos import AbstractPos
from coordinax._src.vectors.collection.core import KinematicSpace

# ===============================================================
# Constructors


@dispatch  # TODO: KinematicSpace[PosT] -- plum#212
def vector(
    cls: type[Coordinate],
    data: KinematicSpace | AbstractPos,
    frame: AbstractReferenceFrame,
    /,
) -> Coordinate:
    """Construct a coordinate from data and a frame.

    Examples
    --------
    >>> import coordinax as cx

    >>> data = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
    >>> cx.Coordinate.from_(data, cx.frames.ICRS())
    Coordinate(
        KinematicSpace({ 'length': CartesianPos3D( ... ) }),
        frame=ICRS()
    )

    """
    return cls(data=data, frame=frame)


@dispatch  # TODO: KinematicSpace[PosT] -- plum#212
def vector(
    cls: type[Coordinate],
    data: KinematicSpace | AbstractPos,
    base_frame: AbstractReferenceFrame,
    ops: AbstractOperator,
    /,
) -> Coordinate:
    """Construct a coordinate from data and a frame.

    Examples
    --------
    >>> import coordinax as cx

    >>> data = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
    >>> cx.Coordinate.from_(data, cx.frames.ICRS(), cx.ops.Identity())
    Coordinate(
        KinematicSpace({ 'length': CartesianPos3D( ... ) }),
        frame=TransformedReferenceFrame(base_frame=ICRS(), xop=Identity())
    )

    """
    frame = TransformedReferenceFrame(base_frame, ops)
    return cls(data=data, frame=frame)


# ===============================================================
# Vector conversion


@dispatch
def vconvert(target: type[AbstractPos], w: Coordinate, /) -> Coordinate:
    """Transform the vector representation of a coordinate.

    Examples
    --------
    >>> import coordinax as cx

    >>> frame = cx.frames.NoFrame()
    >>> data = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
    >>> w = cx.Coordinate(data, frame)

    >>> cx.vconvert(cx.SphericalPos, w)
    Coordinate(
        KinematicSpace({ 'length': SphericalPos( ... ) }),
        frame=NoFrame()
    )

    """
    return replace(w, data=w.data.vconvert(target))



"""Register primitives for AbstractVel."""

__all__: tuple[str, ...] = ()

from typing import cast

import jax
from quax import register

import quaxed.numpy as jnp
import unxt as u
from dataclassish import field_items

from .core import AbstractVel
from coordinax._src.vectors.base_pos import AbstractPos

# ---------------------------------------------------------


@register(jax.lax.mul_p)
def _mul_vel_q(self: AbstractVel, other: u.Quantity["time"]) -> AbstractPos:
    """Multiply the vector by a time `unxt.Quantity` to get a position.

    Examples
    --------
    >>> import quaxed.lax as qlax
    >>> import unxt as u
    >>> import coordinax as cx

    >>> dr = cx.vecs.RadialVel(u.Quantity(1, "m/s"))
    >>> vec = dr * u.Quantity(2, "s")
    >>> print(vec)
    <RadialPos: (r) [m]
        [2]>

    >>> print(qlax.mul(dr, u.Quantity(2, "s")))
    <RadialPos: (r) [m]
        [2]>

    """
    fs = {k: v * other for k, v in field_items(self)}
    return cast("AbstractPos", self.time_antiderivative_cls.from_(fs))


# -----------------------------------------------


@register(jax.lax.neg_p)
def neg_vel(vec: AbstractVel, /) -> AbstractVel:
    """Negate the vector.

    Examples
    --------
    >>> from quaxed import lax
    >>> import unxt as u
    >>> import coordinax as cx

    >>> dr = cx.vecs.RadialVel.from_([1], "m/s")
    >>> print(lax.neg(dr))
    <RadialVel: (r) [m / s]
        [-1]>

    >>> -dr
    RadialVel(r=Quantity(-1, unit='m / s'))

    >>> dp = cx.vecs.PolarVel(u.Quantity(1, "m/s"), u.Quantity(1, "mas/yr"))
    >>> neg_dp = -dp
    >>> print(neg_dp)
    <PolarVel: (r[m / s], phi[mas / yr])
        [-1 -1]>

    """
    return jax.tree.map(jnp.negative, vec)



"""Base classes for operators on coordinates and potentials."""

__all__ = ("AbstractOperator",)

from abc import abstractmethod
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Final

import equinox as eqx
import jax.numpy as jnp
import wadler_lindig as wl
from jaxtyping import ArrayLike
from plum import dispatch

import unxt as u
from dataclassish import field_items, flags

from .api import simplify_op
from coordinax._src.vectors.base_pos import AbstractPos

if TYPE_CHECKING:
    import coordinax.ops  # noqa: ICN001

_sentinel: Final = object()


class AbstractOperator(eqx.Module):
    """Abstract base class for operators on coordinates and potentials.

    An operator is an object that defines a transformation on coordinates. It
    can be applied to a set of coordinates to produce a new set of coordinates.
    Operators can be composed together to form a sequence of transformations.

    When defining a new operator, it should be able to work on the following
    types with the following semantics:

    - `coordinax.vecs.AbstractVector` objects. No restrictions are placed on the
      these arguments, except that if the operator works on
      `coordinax.vecs.AbstractPos` and `coordinax.vecs.AbstractVel` objects, the
      former should precede the latter.
    - Time arguments come first, followed by spatial arguments.
    - It should work on a set of `jax.Array` objects. If the `jax.Array` is a
      coordinate vector then it is Cartesian and the last axis is the component
      axis and prior axes are the batch axes.
    - It should work on `unxt.Quantity` objects.

    """

    # ---------------------------------------------------------------
    # Constructors

    @classmethod
    @dispatch.abstract
    def from_(
        cls: "type[AbstractOperator]", *args: object, **kwargs: object
    ) -> "AbstractOperator":
        """Construct from a set of arguments."""
        raise NotImplementedError  # pragma: no cover

    # ===============================================================
    # Operator API

    @dispatch.abstract
    def __call__(self: "AbstractOperator", *args: Any, **kwargs: Any) -> AbstractPos:
        """Apply the operator to the coordinates `x`."""
        raise NotImplementedError  # pragma: no cover

    # -------------------------------------------

    @property
    @abstractmethod
    def is_inertial(self) -> bool:
        """Whether the operation maintains an inertial reference frame."""
        ...

    @property
    @abstractmethod
    def inverse(self) -> "AbstractOperator":
        """The inverse of the operator."""
        ...

    def simplify(self) -> "AbstractOperator":
        """Simplify the operator.

        This method calls `coordinax.ops.simplify_op` to simplify the
        operator.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax.ops as cxo

        >>> op = cxo.Identity()
        >>> op.simplify() is op
        True

        >>> op = cxo.Pipe((cxo.Identity(), cxo.Identity()))
        >>> op.simplify()
        Identity()

        >>> op = cxo.GalileanOperator(translation=u.Quantity([0., 2., 3., 4.], "km"))
        >>> op.simplify()
        GalileanTranslation(
            delta_t=Quantity(f32[], unit='s'),
            delta_q=CartesianPos3D( ... )
        )

        """
        return simplify_op(self)

    # ===============================================================
    # Wadler-Lindig API

    def __pdoc__(self, **kwargs: Any) -> wl.AbstractDoc:
        """Return the documentation for the operator."""
        # Get the field items, excluding those that should not be shown and
        # those that are equal to the default value.
        # TODO: better filtering using `fields`
        fitems = list(field_items(flags.FilterRepr, self))
        fitems = [
            (k, v)
            for k, v in fitems
            if not jnp.all(v == getattr(self.__class__, k, _sentinel))
        ]
        # Make the field docs list
        if len(fitems) == 1 and kwargs.get("short_arrays") == "compact":
            docs = [wl.TextDoc(str(fitems[0][1]))]
        else:
            docs = wl.named_objs(fitems, **kwargs)

        return wl.bracketed(
            begin=wl.TextDoc(f"{self.__class__.__name__}("),
            docs=docs,
            sep=wl.comma,
            end=wl.TextDoc(")"),
            indent=kwargs.get("indent", 4),
        )

    # ===============================================================
    # Python API

    def __str__(self) -> str:
        """Return a string representation of the operator.

        Examples
        --------
        >>> import jax.numpy as jnp
        >>> import coordinax as cx

        >>> op = cx.ops.GalileanRotation([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        >>> print(op)
        GalileanRotation([[1 0 0]
                          [0 1 0]
                          [0 0 1]])

        >>> op = cx.ops.GalileanOperator(
        ...     translation=u.Quantity([0., 2, 3, 4], "km"),
        ...     velocity=u.Quantity([1., 2, 3], "km/s"),
        ...     rotation=jnp.eye(3).at[0, 2].set(1),
        ... )
        >>> print(op)
        GalileanOperator(
            rotation=GalileanRotation([[1. 0. 1.]
                                       [0. 1. 0.]
                                       [0. 0. 1.]]),
            translation=GalileanTranslation(
                delta_t=Quantity(0., unit='s'),
                delta_q=<CartesianPos3D: (x, y, z) [km]
                    [2. 3. 4.]>
            ),
            velocity=GalileanBoost(<CartesianVel3D: (x, y, z) [km / s]
                    [1. 2. 3.]>)
        )

        """
        return wl.pformat(self, width=88, vector_form=True, short_arrays="compact")

    # ===========================================
    # Operator Composition

    def __or__(self, other: "AbstractOperator") -> "coordinax.ops.Pipe":
        """Compose with another operator.

        Examples
        --------
        >>> import coordinax.ops as cxo

        >>> op1 = cxo.GalileanRotation([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        >>> op2 = cxo.Identity()
        >>> op3 = op1 | op2
        >>> op3
        Pipe((GalileanRotation(rotation=i32[3,3]), Identity()))

        >>> op4 = cxo.Identity() | op3
        >>> op4
        Pipe((Identity(), GalileanRotation(rotation=i32[3,3]), Identity()))

        """
        from .pipe import Pipe  # noqa: PLC0415

        if isinstance(other, Pipe):
            return other.__ror__(self)
        return Pipe((self, other))


# ============================================================
# Constructors


@AbstractOperator.from_.dispatch(precedence=-1)
def from_(
    cls: type[AbstractOperator], *args: object, **kwargs: object
) -> AbstractOperator:
    """Construct from a set of arguments.

    This is a low-priority dispatch that will be called if no other
    dispatch is found. It just tries to pass the arguments to the
    constructor.

    """
    return cls(*args, **kwargs)


@AbstractOperator.from_.dispatch
def from_(cls: type[AbstractOperator], obj: Mapping[str, Any], /) -> AbstractOperator:
    """Construct from a mapping.

    Examples
    --------
    >>> import coordinax.ops as cxo
    >>> pipe = cxo.Identity() | cxo.Identity()
    >>> cxo.Pipe.from_({"operators": pipe})
    Pipe((Identity(), Identity()))

    """
    return cls(**obj)


@AbstractOperator.from_.dispatch
def from_(
    cls: type[AbstractOperator],
    x: ArrayLike | list[float | int],
    unit: str,  # TODO: support unit object
    /,
) -> AbstractOperator:
    """Construct from a Quantity's value and unit.

    Examples
    --------
    >>> import coordinax.ops as cxo

    >>> op = cxo.GalileanSpatialTranslation.from_([1, 1, 1], "km")
    >>> print(op)
    GalileanSpatialTranslation(<CartesianPos3D: (x, y, z) [km]
        [1 1 1]>)

    >>> op = cxo.GalileanTranslation.from_([3e5, 1, 1, 1], "km")
    >>> print(op)
    GalileanTranslation(
      delta_t=Quantity(1.0006922, unit='s'),
      delta_q=<CartesianPos3D: (x, y, z) [km]
          [1. 1. 1.]>
    )

    >>> op = cxo.GalileanBoost.from_([1, 1, 1], "km/s")
    >>> print(op)
    GalileanBoost(<CartesianVel3D: (x, y, z) [km / s]
        [1 1 1]>)

    """
    return cls.from_(u.Quantity(x, unit))


@AbstractOperator.from_.dispatch
def from_(cls: type[AbstractOperator], obj: AbstractOperator, /) -> AbstractOperator:
    """Construct an operator from another operator.

    Raises
    ------
    TypeError
        If the input object is not a subclass of the target class.

    Examples
    --------
    >>> import coordinax as cx

    If the object is the same type, it should return the object itself.

    >>> op = cx.ops.Identity()
    >>> cx.ops.Identity.from_(op) is op
    True

    If the object is a different type, it will error.

    >>> try:
    ...     cx.ops.GalileanBoost.from_(op)
    ... except TypeError as e:
    ...     print(e)
    Cannot construct <class 'coordinax...GalileanBoost'> from <class 'coordinax...Identity'>.

    Unless the object is a subclass of the target class.

    >>> class MyOperator(cx.ops.Identity):
    ...     pass

    >>> op = MyOperator()
    >>> op
    MyOperator()

    >>> newop = cx.ops.Identity.from_(op)
    >>> newop is op, isinstance(newop, cx.ops.Identity)
    (False, True)

    """  # noqa: E501
    if not isinstance(obj, cls):
        msg = f"Cannot construct {cls} from {type(obj)}."
        raise TypeError(msg)

    # avoid copying if the types are the same. Isinstance is not strict
    # enough, so we use type() instead.
    if type(obj) is cls:  # pylint: disable=unidiomatic-typecheck
        return obj

    return cls(**dict(field_items(obj)))



"""`coordinax.distance` private module.

Note that this module is private. Users should use the public API.

This module depends on the following modules:

- utils & typing

"""

__all__ = (
    "AbstractDistance",
    "Distance",
    "DistanceModulus",
    "Parallax",
    # funcs
    "distance",
    "parallax",
    "distance_modulus",
    # Typing
    "BBtLength",
    "BatchableDistance",
)

from .base import AbstractDistance
from .custom_types import BatchableDistance, BBtLength
from .funcs import distance, distance_modulus, parallax
from .measures import Distance, DistanceModulus, Parallax

# isort: split
# Register the dispatching
from . import (  # noqa: F401
    register_constructors,
    register_converters,
    register_primitives,
    register_unxt,
)



"""Register Support with Unxt."""

__all__ = ("ToUnitsOptions",)


from collections.abc import Mapping
from enum import Enum
from typing import Any, Literal

from astropy.units import PhysicalType as Dimension
from plum import dispatch

import unxt as u
from dataclassish import field_items, replace

from .flags import AttrFilter
from .vector import AbstractVector
from coordinax._src.custom_types import Unit


class ToUnitsOptions(Enum):
    """Options for the units argument of `AbstractVector.uconvert`."""

    consistent = "consistent"
    """Convert to consistent units."""


@dispatch
def uconvert(usys: u.AbstractUnitSystem, vector: AbstractVector, /) -> AbstractVector:
    """Convert the vector to the given units.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> usys = u.unitsystem("m", "s", "kg", "rad")

    >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> print(u.uconvert(usys, vec))
    <CartesianPos3D: (x, y, z) [m]
        [1000. 2000. 3000.]>

    """
    usys = u.unitsystem(usys)
    return replace(
        vector,
        **{
            k: u.uconvert(usys[u.dimension_of(v)], v)
            for k, v in field_items(AttrFilter, vector)
        },
    )


@dispatch
def uconvert(
    units: Mapping[Dimension, Unit | str], vector: AbstractVector, /
) -> AbstractVector:
    """Convert the vector to the given units.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    We can convert a vector to the given units:

    >>> cart = cx.vecs.CartesianPos2D(x=u.Quantity(1, "m"), y=u.Quantity(2, "km"))
    >>> cart.uconvert({u.dimension("length"): "km"})
    CartesianPos2D(x=Quantity(0.001, unit='km'), y=Quantity(2, unit='km'))

    This also works for vectors with different units:

    >>> sph = cx.SphericalPos(r=u.Quantity(1, "m"), theta=u.Quantity(45, "deg"),
    ...                       phi=u.Quantity(3, "rad"))
    >>> sph.uconvert({u.dimension("length"): "km", u.dimension("angle"): "deg"})
    SphericalPos(
      r=Distance(0.001, unit='km'),
      theta=Angle(45, unit='deg'),
      phi=Angle(171.88734, unit='deg')
    )

    """
    # Ensure `units_` is PT -> Unit
    units_ = {u.dimension(k): v for k, v in units.items()}
    # Convert to the given units
    return replace(
        vector,
        **{
            k: u.uconvert(units_[u.dimension_of(v)], v)
            for k, v in field_items(AttrFilter, vector)
        },
    )


@dispatch
def uconvert(units: Mapping[str, Any], vector: AbstractVector, /) -> AbstractVector:
    """Convert the vector to the given units.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    We can convert a vector to the given units:

    >>> cart = cx.vecs.CartesianPos2D(x=u.Quantity(1, "m"), y=u.Quantity(2, "km"))
    >>> cart.uconvert({"x": "km", "y": "m"})
    CartesianPos2D(x=Quantity(0.001, unit='km'), y=Quantity(2000., unit='m'))

    This also works for converting just some of the components:

    >>> cart.uconvert({"x": "km"})
    CartesianPos2D(x=Quantity(0.001, unit='km'), y=Quantity(2, unit='km'))

    This also works for vectors with different units:

    >>> sph = cx.SphericalPos(r=u.Quantity(1, "m"), theta=u.Quantity(45, "deg"),
    ...                       phi=u.Quantity(3, "rad"))
    >>> sph.uconvert({"r": "km", "theta": "rad"})
    SphericalPos(
      r=Distance(0.001, unit='km'),
      theta=Angle(0.7853982, unit='rad'),
      phi=Angle(3., unit='rad')
    )

    """
    return replace(
        vector,
        **{
            k: u.uconvert(units.get(k, u.unit_of(v)), v)
            for k, v in field_items(AttrFilter, vector)
        },
    )


@dispatch
def uconvert(
    flag: Literal[ToUnitsOptions.consistent], vector: AbstractVector, /
) -> AbstractVector:
    """Convert the vector to a self-consistent set of units.

    Parameters
    ----------
    flag
        The vector is converted to consistent units by looking for the first
        quantity with each physical type and converting all components to
        the units of that quantity.
    vector
        The vector to convert.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    We can convert a vector to the given units:

    >>> cart = cx.vecs.CartesianPos2D.from_([1, 2], "m")

    If all you want is to convert to consistent units, you can use
    ``"consistent"``:

    >>> cart.uconvert(cx.vecs.ToUnitsOptions.consistent)
    CartesianPos2D(x=Quantity(1, unit='m'), y=Quantity(2, unit='m'))

    >>> sph = cart.vconvert(cx.SphericalPos)
    >>> sph.uconvert(cx.vecs.ToUnitsOptions.consistent)
    SphericalPos(
      r=Distance(2.236068, unit='m'),
      theta=Angle(1.5707964, unit='rad'),
      phi=Angle(1.1071488, unit='rad')
    )

    """
    dim2unit = {}
    units_ = {}
    for k, v in field_items(AttrFilter, vector):
        pt = u.dimension_of(v)
        if pt not in dim2unit:
            dim2unit[pt] = u.unit_of(v)
        units_[k] = dim2unit[pt]

    return replace(
        vector,
        **{k: u.uconvert(units_[k], v) for k, v in field_items(AttrFilter, vector)},
    )


@dispatch
def uconvert(usys: str, vector: AbstractVector, /) -> AbstractVector:
    """Convert the vector to the given units system.

    Parameters
    ----------
    usys
        The units system to convert to, as a string.
    vector
        The vector to convert.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> usys = "galactic"
    >>> vector = cx.CartesianPos3D.from_([1, 2, 3], "m")
    >>> u.uconvert(usys, vector)
    CartesianPos3D(
      x=Quantity(3.2407793e-20, unit='kpc'),
      y=Quantity(6.4815585e-20, unit='kpc'),
      z=Quantity(9.722338e-20, unit='kpc')
    )

    """
    usys = u.unitsystem(usys)
    return uconvert(usys, vector)



"""Transformations between representations."""

__all__: tuple[str, ...] = ()


from typing import TYPE_CHECKING, cast

from plum import dispatch

import quaxed.numpy as jnp

from coordinax._src.vectors.collection import KinematicSpace
from coordinax._src.vectors.d3 import CylindricalPos, CylindricalVel
from coordinax._src.vectors.dn import PoincarePolarVector

if TYPE_CHECKING:
    import unxt  # noqa: ICN001


@dispatch  # TODO: KinematicSpace[CylindricalPos] -- plum#212
def vconvert(target: type[KinematicSpace], w: PoincarePolarVector, /) -> KinematicSpace:
    """Space -> PoincarePolarVector.

    Examples
    --------
    >>> import coordinax as cx

    >>> w = cx.KinematicSpace(
    ...     length=cx.CartesianPos3D.from_([[[1, 2, 3], [4, 5, 6]]], "m"),
    ...     speed=cx.CartesianVel3D.from_([[[1, 2, 3], [4, 5, 6]]], "m/s")
    ... )

    >>> cx.vconvert(cx.vecs.PoincarePolarVector, w)
    PoincarePolarVector(
      rho=Quantity([[2.23606801, 6.40312433]], unit='m'),
      pp_phi=Quantity([[0., 0.]], unit='m / s(1/2)'),
      z=Quantity([[3, 6]], unit='m'),
      dt_rho=Quantity([[2.23606801, 6.40312433]], unit='m / s'),
      dt_pp_phi=Quantity([[0., 0.]], unit='m / s(1/2)'),
      dt_z=Quantity([[3, 6]], unit='m / s')
    )

    >>> cx.vconvert(cx.KinematicSpace, w)
    KinematicSpace({
      'length': CartesianPos3D(
        x=Quantity([[1, 4]], unit='m'),
        y=Quantity([[2, 5]], unit='m'),
        z=Quantity([[3, 6]], unit='m')
      ),
      'speed': CartesianVel3D(
        x=Quantity([[1, 4]], unit='m / s'),
        y=Quantity([[2, 5]], unit='m / s'),
        z=Quantity([[3, 6]], unit='m / s')
      )
    })

    """
    phi = cast("unxt.AbstractQuantity", jnp.atan2(w.dt_pp_phi, w.pp_phi))
    dt_phi = (w.pp_phi**2 + w.dt_pp_phi**2) / 2 / w.rho**2  # TODO: note the abs

    return KinematicSpace(
        length=CylindricalPos(rho=w.rho, z=w.z, phi=phi),
        speed=CylindricalVel(rho=w.dt_rho, z=w.dt_z, phi=dt_phi),
    )



"""Abstract Accelerations."""

__all__ = ("AbstractAcc", "ACCELERATION_CLASSES")

import functools as ft
from typing import TYPE_CHECKING, Any, cast

import jax

import unxt as u

from coordinax._src.utils import classproperty
from coordinax._src.vectors import api
from coordinax._src.vectors.base import AbstractVector
from coordinax._src.vectors.base_pos import AbstractPos
from coordinax._src.vectors.base_vel import AbstractVel
from coordinax._src.vectors.mixins import AvalMixin

if TYPE_CHECKING:
    import coordinax.vecs  # noqa: ICN001


class AbstractAcc(AvalMixin, AbstractVector):  # pylint: disable=abstract-method
    """Abstract representation of vector differentials in different systems."""

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Initialize the subclass.

        The subclass is registered.
        """
        super().__init_subclass__(**kwargs)

        ACCELERATION_CLASSES_MUTABLE[cls] = None

    # ===============================================================
    # Vector API

    @classproperty
    @classmethod
    def cartesian_type(cls) -> "type[coordinax.vecs.AbstractAcc]":
        """Return the corresponding Cartesian vector class."""
        return api.cartesian_vector_type(cls)

    @classproperty
    @classmethod
    def time_derivative_cls(cls) -> "type[coordinax.vecs.AbstractVector]":
        """Return the corresponding time derivative class."""
        return api.time_derivative_vector_type(cls)

    @classproperty
    @classmethod
    def time_antiderivative_cls(cls) -> type[AbstractVel]:
        """Return the corresponding time antiderivative class."""
        return api.time_antiderivative_vector_type(cls)

    @classmethod
    def time_nth_derivative_cls(cls, n: int) -> "type[coordinax.vecs.AbstractVector]":
        """Return the corresponding time nth derivative class."""
        return api.time_nth_derivative_vector_type(cls, n=n)

    # ===============================================================
    # Convenience methods

    @ft.partial(jax.jit)
    def norm(
        self: "AbstractAcc", p: AbstractVel, q: AbstractPos, /
    ) -> u.Quantity["acceleration"]:
        """Return the norm of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> q = cx.vecs.CartesianPos3D.from_([1, 2, 3], "kpc")
        >>> p = cx.vecs.CartesianVel3D.from_([4, 5, 6], "km/s")
        >>> a = cx.vecs.CartesianAcc3D.from_([3, 4, 0], "m/s2")

        >>> a = a.vconvert(cx.vecs.CylindricalAcc, p, q)

        >>> a.norm(p, q)
        Quantity(Array(5..., dtype=float32), unit='m / s2')

        """
        cart_cls = self.cartesian_type
        cart_acc = cast("AbstractAcc", self.vconvert(cart_cls, p, q))
        return cart_acc.norm(p, q)


# -----------------

ACCELERATION_CLASSES_MUTABLE: dict[type[AbstractAcc], None] = {}
ACCELERATION_CLASSES = ACCELERATION_CLASSES_MUTABLE.keys()



"""Coordinax package."""

from plum import add_promotion_rule

import unxt as u

from .distances import AbstractDistance

# Add a rule that when a AbstractDistance interacts with a Quantity, the
# distance degrades to a Quantity. This is necessary for many operations, e.g.
# division of a distance by non-dimensionless quantity where the resulting units
# are not those of a distance.
add_promotion_rule(AbstractDistance, u.quantity.AbstractAngle, u.Quantity)
